# ruff: noqa: F401
from .cifar import CIFAR10C, CIFAR10H, CIFAR10N, CIFAR100C, CIFAR100N
from .cub import CUB
from .imagenet import (
    ImageNetA,
    ImageNetC,
    ImageNetO,
    ImageNetR,
    TinyImageNet,
    TinyImageNetC,
)
from .mnist_c import MNISTC
from .not_mnist import NotMNIST
from .openimage_o import OpenImageO
from .uci import (
    HTRU2,
    BankMarketing,
    DOTA2Games,
    OnlineShoppers,
    SpamBase,
)



import torch
from torch import Tensor, nn

from torch_uncertainty.layers.bayesian import bayesian_modules


class StochasticModel(nn.Module):
    def __init__(self, model: nn.Module, num_samples: int) -> None:
        super().__init__()
        self.core_model = model
        self.num_samples = num_samples

    def eval_forward(self, x: Tensor) -> Tensor:
        return torch.cat([self.core_model.forward(x) for _ in range(self.num_samples)], dim=0)

    def forward(self, x: Tensor) -> Tensor:
        if self.training:
            return self.core_model.forward(x)
        return self.eval_forward(x)

    def sample(self, num_samples: int = 1) -> list[dict]:
        sampled_models = [{}] * num_samples
        for module_name in self.core_model._modules:
            module = self.core_model._modules[module_name]
            if isinstance(module, bayesian_modules):
                for model in sampled_models:
                    weight, bias = module.sample()
                    model[module_name + ".weight"] = weight
                    if bias is not None:
                        model[module_name + ".bias"] = bias
            else:
                for model in sampled_models:
                    state = module.state_dict()
                    if not len(state):  # no parameter
                        break
                    # TODO: fix this
                    model |= {
                        module_name + "." + key: val for key, val in module.state_dict().items()
                    }
        return sampled_models

    def freeze(self) -> None:
        for module_name in self.core_model._modules:
            module = self.core_model._modules[module_name]
            if isinstance(module, bayesian_modules):
                module.freeze()

    def unfreeze(self) -> None:
        for module_name in self.core_model._modules:
            module = self.core_model._modules[module_name]
            if isinstance(module, bayesian_modules):
                module.unfreeze()



import torch
from einops import rearrange
from lightning.pytorch import LightningModule
from lightning.pytorch.loggers import Logger
from lightning.pytorch.utilities.types import STEP_OUTPUT
from torch import Tensor, nn
from torch.optim import Optimizer
from torchmetrics import Accuracy, MetricCollection
from torchvision.transforms.v2 import ToDtype
from torchvision.transforms.v2 import functional as F
from torchvision.utils import draw_segmentation_masks

from torch_uncertainty.metrics import (
    AUGRC,
    AURC,
    BrierScore,
    CalibrationError,
    CategoricalNLL,
    MeanIntersectionOverUnion,
)
from torch_uncertainty.models import (
    EPOCH_UPDATE_MODEL,
    STEP_UPDATE_MODEL,
)
from torch_uncertainty.utils.plotting import show


class SegmentationRoutine(LightningModule):
    def __init__(
        self,
        model: nn.Module,
        num_classes: int,
        loss: nn.Module,
        optim_recipe: dict | Optimizer | None = None,
        eval_shift: bool = False,
        format_batch_fn: nn.Module | None = None,
        metric_subsampling_rate: float = 1e-2,
        log_plots: bool = False,
        num_samples_to_plot: int = 3,
        num_calibration_bins: int = 15,
    ) -> None:
        r"""Routine for training & testing on **segmentation** tasks.

        Args:
            model (torch.nn.Module): Model to train.
            num_classes (int): Number of classes in the segmentation task.
            loss (torch.nn.Module): Loss function to optimize the :attr:`model`.
            optim_recipe (dict or Optimizer, optional): The optimizer and
                optionally the scheduler to use. Defaults to ``None``.
            eval_shift (bool, optional): Indicates whether to evaluate the Distribution
                shift performance. Defaults to ``False``.
            format_batch_fn (torch.nn.Module, optional): The function to format the
                batch. Defaults to ``None``.
            metric_subsampling_rate (float, optional): The rate of subsampling for the
                memory consuming metrics. Defaults to ``1e-2``.
            log_plots (bool, optional): Indicates whether to log plots from
                metrics. Defaults to ``False``.
            num_samples_to_plot (int, optional): Number of samples to plot in the
                segmentation results. Defaults to ``3``.
            num_calibration_bins (int, optional): Number of bins to compute calibration
                metrics. Defaults to ``15``.

        Warning:
            You must define :attr:`optim_recipe` if you do not use the CLI.

        Note:
            :attr:`optim_recipe` can be anything that can be returned by
            :meth:`LightningModule.configure_optimizers()`. Find more details
            `here <https://lightning.ai/docs/pytorch/stable/common/lightning_module.html#configure-optimizers>`_.
        """
        super().__init__()
        _segmentation_routine_checks(
            num_classes=num_classes,
            metric_subsampling_rate=metric_subsampling_rate,
            num_calibration_bins=num_calibration_bins,
        )
        if eval_shift:
            raise NotImplementedError(
                "Distribution shift evaluation not implemented yet. Raise an issue if needed."
            )

        self.model = model
        self.num_classes = num_classes
        self.num_calibration_bins = num_calibration_bins
        self.loss = loss
        self.needs_epoch_update = isinstance(model, EPOCH_UPDATE_MODEL)
        self.needs_step_update = isinstance(model, STEP_UPDATE_MODEL)

        if format_batch_fn is None:
            format_batch_fn = nn.Identity()

        self.optim_recipe = optim_recipe
        self.format_batch_fn = format_batch_fn
        self.metric_subsampling_rate = metric_subsampling_rate
        self.log_plots = log_plots
        self._init_metrics()

        if log_plots:
            self.num_samples_to_plot = num_samples_to_plot
            self.sample_buffer = []

    def _init_metrics(self) -> None:
        """Initialize the metrics depending on the exact task."""
        seg_metrics = MetricCollection(
            {
                "seg/mIoU": MeanIntersectionOverUnion(num_classes=self.num_classes),
            },
            compute_groups=False,
        )
        sbsmpl_seg_metrics = MetricCollection(
            {
                "seg/mAcc": Accuracy(
                    task="multiclass", average="macro", num_classes=self.num_classes
                ),
                "seg/Brier": BrierScore(num_classes=self.num_classes),
                "seg/NLL": CategoricalNLL(),
                "seg/pixAcc": Accuracy(task="multiclass", num_classes=self.num_classes),
                "cal/ECE": CalibrationError(
                    task="multiclass",
                    num_classes=self.num_classes,
                    num_bins=self.num_calibration_bins,
                ),
                "cal/aECE": CalibrationError(
                    task="multiclass",
                    adaptive=True,
                    num_classes=self.num_classes,
                    num_bins=self.num_calibration_bins,
                ),
                "sc/AURC": AURC(),
                "sc/AUGRC": AUGRC(),
            },
            compute_groups=[
                ["seg/mAcc"],
                ["seg/Brier"],
                ["seg/NLL"],
                ["seg/pixAcc"],
                ["cal/ECE", "cal/aECE"],
                ["sc/AURC", "sc/AUGRC"],
            ],
        )

        self.val_seg_metrics = seg_metrics.clone(prefix="val/")
        self.val_sbsmpl_seg_metrics = sbsmpl_seg_metrics.clone(prefix="val/")
        self.test_seg_metrics = seg_metrics.clone(prefix="test/")
        self.test_sbsmpl_seg_metrics = sbsmpl_seg_metrics.clone(prefix="test/")

    def configure_optimizers(self) -> Optimizer | dict:
        return self.optim_recipe

    def forward(self, inputs: Tensor) -> Tensor:
        """Forward pass of the model.

        Args:
            inputs (Tensor): input tensor.

        Returns:
            Tensor: the prediction of the model.
        """
        return self.model(inputs)

    def on_train_start(self) -> None:
        if self.logger is not None:  # coverage: ignore
            self.logger.log_hyperparams(self.hparams)

    def on_validation_start(self) -> None:
        if self.needs_epoch_update and not self.trainer.sanity_checking:
            self.model.update_wrapper(self.current_epoch)
            if hasattr(self.model, "need_bn_update"):
                self.model.bn_update(self.trainer.train_dataloader, device=self.device)

    def on_test_start(self) -> None:
        if hasattr(self.model, "need_bn_update"):
            self.model.bn_update(self.trainer.train_dataloader, device=self.device)

    def training_step(self, batch: tuple[Tensor, Tensor]) -> STEP_OUTPUT:
        """Perform a single training step based on the input tensors.

        Args:
            batch (tuple[Tensor, Tensor]): the training images and their corresponding targets

        Returns:
            Tensor: the loss corresponding to this training step.
        """
        img, target = self.format_batch_fn(batch)
        logits = self.forward(img)
        target = F.resize(target, logits.shape[-2:], interpolation=F.InterpolationMode.NEAREST)
        logits = rearrange(logits, "b c h w -> (b h w) c")
        target = target.flatten()
        valid_mask = target != 255
        loss = self.loss(logits[valid_mask], target[valid_mask])
        if self.needs_step_update:
            self.model.update_wrapper(self.current_epoch)
        self.log("train_loss", loss, prog_bar=True, logger=True)
        return loss

    def validation_step(self, batch: tuple[Tensor, Tensor]) -> None:
        """Perform a single validation step based on the input tensors.

        Compute the prediction of the model and the value of the metrics on the validation batch.

        Args:
            batch (tuple[Tensor, Tensor]): the validation images and their corresponding targets
        """
        img, targets = batch
        logits = self.forward(img)
        targets = F.resize(
            targets,
            logits.shape[-2:],
            interpolation=F.InterpolationMode.NEAREST,
        )
        logits = rearrange(logits, "(m b) c h w -> (b h w) m c", b=targets.size(0))
        probs_per_est = logits.softmax(dim=-1)
        probs = probs_per_est.mean(dim=1)
        targets = targets.flatten()
        valid_mask = targets != 255
        probs, targets = probs[valid_mask], targets[valid_mask]
        self.val_seg_metrics.update(probs, targets)
        self.val_sbsmpl_seg_metrics.update(*self.subsample(probs, targets))

    def test_step(self, batch: tuple[Tensor, Tensor]) -> None:
        """Perform a single test step based on the input tensors.

        Compute the prediction of the model and the value of the metrics on the test batch.

        Args:
            batch (tuple[Tensor, Tensor]): the test images and their corresponding targets
        """
        img, targets = batch
        logits = self.forward(img)
        targets = F.resize(
            targets,
            logits.shape[-2:],
            interpolation=F.InterpolationMode.NEAREST,
        )

        logits = rearrange(logits, "(m b) c h w -> b m c h w", b=targets.size(0))
        probs_per_est = logits.softmax(dim=2)
        probs = probs_per_est.mean(dim=1)

        if self.log_plots and len(self.sample_buffer) < self.num_samples_to_plot:
            max_count = self.num_samples_to_plot - len(self.sample_buffer)
            for i, (_img, _prb, _tgt) in enumerate(zip(img, probs, targets, strict=False)):
                if i >= max_count:
                    break
                _pred = _prb.argmax(dim=0, keepdim=True)
                self.sample_buffer.append((_img, _pred, _tgt))

        probs = rearrange(probs, "b c h w -> (b h w) c")
        targets = targets.flatten()
        valid_mask = targets != 255
        probs, targets = probs[valid_mask], targets[valid_mask]
        self.test_seg_metrics.update(probs, targets)
        self.test_sbsmpl_seg_metrics.update(*self.subsample(probs, targets))

    def on_validation_epoch_end(self) -> None:
        """Compute and log the values of the collected metrics in `validation_step`."""
        res_dict = self.val_seg_metrics.compute()
        self.log_dict(res_dict, logger=True, sync_dist=True)
        self.log(
            "mIoU%",
            res_dict["val/seg/mIoU"] * 100,
            prog_bar=True,
            sync_dist=True,
        )
        self.log_dict(self.val_sbsmpl_seg_metrics.compute(), sync_dist=True)
        self.val_seg_metrics.reset()
        self.val_sbsmpl_seg_metrics.reset()

    def on_test_epoch_end(self) -> None:
        """Compute, log, and plot the values of the collected metrics in `test_step`."""
        self.log_dict(self.test_seg_metrics.compute(), sync_dist=True)
        self.log_dict(self.test_sbsmpl_seg_metrics.compute(), sync_dist=True)
        if isinstance(self.logger, Logger) and self.log_plots:
            self.logger.experiment.add_figure(
                "Calibration/Reliabity diagram",
                self.test_sbsmpl_seg_metrics["cal/ECE"].plot()[0],
            )
            self.logger.experiment.add_figure(
                "Selective Classification/Risk-Coverage curve",
                self.test_sbsmpl_seg_metrics["sc/AURC"].plot()[0],
            )
            self.logger.experiment.add_figure(
                "Selective Classification/Generalized Risk-Coverage curve",
                self.test_sbsmpl_seg_metrics["sc/AUGRC"].plot()[0],
            )
            self.log_segmentation_plots()

    def log_segmentation_plots(self) -> None:
        """Build and log examples of segmentation plots from the test set."""
        for i, (img, pred, tgt) in enumerate(self.sample_buffer):
            pred = pred == torch.arange(self.num_classes, device=pred.device)[:, None, None]
            tgt = tgt == torch.arange(self.num_classes, device=tgt.device)[:, None, None]

            # Undo normalization on the image and convert to uint8.
            mean = torch.tensor(self.trainer.datamodule.mean, device=img.device)
            std = torch.tensor(self.trainer.datamodule.std, device=img.device)
            img = img * std[:, None, None] + mean[:, None, None]
            img = ToDtype(torch.uint8, scale=True)(img)

            dataset = self.trainer.datamodule.test
            color_palette = dataset.color_palette if hasattr(dataset, "color_palette") else None

            pred_mask = draw_segmentation_masks(img, pred, alpha=0.7, colors=color_palette)
            gt_mask = draw_segmentation_masks(img, tgt, alpha=0.7, colors=color_palette)

            self.logger.experiment.add_figure(
                f"Segmentation results/{i}",
                show(pred_mask, gt_mask),
            )

    def subsample(self, pred: Tensor, target: Tensor) -> tuple[Tensor, Tensor]:
        """Select a random sample of the data to compute the loss onto.

        Args:
            pred (Tensor): the prediction tensor.
            target (Tensor): the target tensor.

        Returns:
            Tuple[Tensor, Tensor]: the subsampled prediction and target tensors.
        """
        total_size = target.size(0)
        num_samples = max(1, int(total_size * self.metric_subsampling_rate))
        indices = torch.randperm(total_size, device=pred.device)[:num_samples]
        return pred[indices], target[indices]


def _segmentation_routine_checks(
    num_classes: int,
    metric_subsampling_rate: float,
    num_calibration_bins: int,
) -> None:
    """Check the domains of the routine's parameters.

    Args:
        num_classes (int): the number of classes in the dataset.
        metric_subsampling_rate (float): the rate of subsampling to compute the metrics.
        num_calibration_bins (int): the number of bins for the evaluation of the calibration.
    """
    if num_classes < 2:
        raise ValueError(f"num_classes must be at least 2, got {num_classes}.")

    if not 0 < metric_subsampling_rate <= 1:
        raise ValueError(
            f"metric_subsampling_rate must be in the range (0, 1], got {metric_subsampling_rate}."
        )

    if num_calibration_bins < 2:
        raise ValueError(f"num_calibration_bins must be at least 2, got {num_calibration_bins}.")



from typing import Literal

from torch import nn

from torch_uncertainty.models.segmentation.segformer import (
    seg_former,
)
from torch_uncertainty.routines.segmentation import SegmentationRoutine


class SegFormerBaseline(SegmentationRoutine):
    single = ["std"]
    versions = {
        "std": seg_former,
    }
    archs = [0, 1, 2, 3, 4, 5]

    def __init__(
        self,
        num_classes: int,
        loss: nn.Module,
        version: Literal["std"],
        arch: int,
    ) -> None:
        r"""SegFormer backbone baseline for segmentation providing support for
        various versions and architectures.

        Args:
            num_classes (int): Number of classes to predict.
            loss (type[Module]): Training loss.
            version (str):
                Determines which SegFormer version to use. Options are:

                - ``"std"``: original SegFormer

            arch (int):
                Determines which architecture to use. Options are:

                - ``0``: SegFormer-B0
                - ``1``: SegFormer-B1
                - ``2``: SegFormer-B2
                - ``3``: SegFormer-B3
                - ``4``: SegFormer-B4
                - ``5``: SegFormer-B5

            num_estimators (int, optional): Number of estimators in the
                ensemble. Defaults to 1 (single model).
        """
        params = {
            "num_classes": num_classes,
            "arch": arch,
        }

        format_batch_fn = nn.Identity()

        if version not in self.versions:
            raise ValueError(f"Unknown version {version}")

        model = self.versions[version](**params)

        super().__init__(
            num_classes=num_classes,
            model=model,
            loss=loss,
            format_batch_fn=format_batch_fn,
        )
        self.save_hyperparameters(ignore=["loss"])



import math

import pytest
import torch

from torch_uncertainty.metrics import MutualInformation


@pytest.fixture()
def disagreement_probas() -> torch.Tensor:
    """Return a vector with mean entropy ~ln(2) and entropy of mean =0."""
    return torch.as_tensor([[[1e-8, 1 - 1e-8], [1 - 1e-8, 1e-8]]])


@pytest.fixture()
def agreement_probas() -> torch.Tensor:
    return torch.as_tensor([[[0.9, 0.1], [0.9, 0.1]]])


class TestMutualInformation:
    """Testing the MutualInformation metric class."""

    def test_compute_disagreement(self, disagreement_probas: torch.Tensor):
        metric = MutualInformation()
        metric.update(disagreement_probas)
        res = metric.compute()
        assert res == pytest.approx(math.log(2), 1e-5)

    def test_compute_agreement(self, agreement_probas: torch.Tensor):
        metric = MutualInformation()
        metric.update(agreement_probas)
        res = metric.compute()
        assert res == 0.0

    def test_compute_mixed(self, disagreement_probas: torch.Tensor, agreement_probas: torch.Tensor):
        metric = MutualInformation(reduction="mean")
        metric.update(agreement_probas)
        metric.update(disagreement_probas)
        res = metric.compute()
        assert res == pytest.approx(math.log(2) / 2, 1e-5)

        metric = MutualInformation(reduction="sum")
        metric.update(agreement_probas)
        metric.update(disagreement_probas)
        res = metric.compute()
        assert res == pytest.approx(math.log(2), 1e-5)

        metric = MutualInformation(reduction=None)
        metric.update(agreement_probas)
        metric.update(disagreement_probas)
        res = metric.compute()
        assert res[0] == pytest.approx(0, 1e-5)
        assert res[1] == pytest.approx(math.log(2), 1e-5)

    def test_bad_argument(self):
        with pytest.raises(ValueError):
            _ = MutualInformation("geometric_mean")



from typing import Any, Literal

import torch
from torch import Tensor
from torch.nn.utils.rnn import pad_sequence
from torchmetrics.classification.calibration_error import (
    _binary_calibration_error_arg_validation,
    _multiclass_calibration_error_arg_validation,
)
from torchmetrics.metric import Metric
from torchmetrics.utilities.data import dim_zero_cat
from torchmetrics.utilities.enums import ClassificationTaskNoMultilabel


def _equal_binning_bucketize(
    confidences: Tensor, accuracies: Tensor, num_bins: int
) -> tuple[Tensor, Tensor, Tensor]:
    """Compute bins for the adaptive calibration error.

    Args:
        confidences: The confidence (i.e. predicted prob) of the top1
            prediction.
        accuracies: 1.0 if the top-1 prediction was correct, 0.0 otherwise.
        num_bins: Number of bins to use when computing adaptive calibration
            error.

    Returns:
        tuple with binned accuracy, binned confidence and binned probabilities
    """
    confidences, indices = torch.sort(confidences)
    accuracies = accuracies[indices]
    acc_bin, conf_bin = (
        accuracies.tensor_split(num_bins),
        confidences.tensor_split(num_bins),
    )
    count_bin = torch.as_tensor(
        [len(cb) for cb in conf_bin],
        dtype=confidences.dtype,
        device=confidences.device,
    )
    return (
        pad_sequence(acc_bin, batch_first=True).sum(1) / count_bin,
        pad_sequence(conf_bin, batch_first=True).sum(1) / count_bin,
        torch.as_tensor(count_bin) / len(confidences),
    )


def _ace_compute(
    confidences: Tensor,
    accuracies: Tensor,
    num_bins: int,
    norm: Literal["l1", "l2", "max"] = "l1",
    debias: bool = False,
) -> Tensor:
    """Compute the adaptive calibration error given the provided number of bins
        and norm.

    Args:
        confidences: The confidence (i.e. predicted prob) of the top1
            prediction.
        accuracies: 1.0 if the top-1 prediction was correct, 0.0 otherwise.
        num_bins: Number of bins to use when computing adaptive calibration
            error.
        norm: Norm function to use when computing calibration error. Defaults
            to "l1".
        debias: Apply debiasing to L2 norm computation as in
            `Verified Uncertainty Calibration`. Defaults to False.

    Returns:
        Tensor: Adaptive Calibration error scalar.
    """
    with torch.no_grad():
        acc_bin, conf_bin, prop_bin = _equal_binning_bucketize(confidences, accuracies, num_bins)

    if norm == "l1":
        return torch.sum(torch.abs(acc_bin - conf_bin) * prop_bin)
    if norm == "max":
        ace = torch.max(torch.abs(acc_bin - conf_bin))
    if norm == "l2":
        ace = torch.sum(torch.pow(acc_bin - conf_bin, 2) * prop_bin)
        if debias:  # coverage: ignore
            debias_bins = (acc_bin * (acc_bin - 1) * prop_bin) / (
                prop_bin * accuracies.size()[0] - 1
            )
            ace += torch.sum(
                torch.nan_to_num(debias_bins)
            )  # replace nans with zeros if nothing appeared in a bin
        return torch.sqrt(ace) if ace > 0 else torch.tensor(0)
    return ace


class BinaryAdaptiveCalibrationError(Metric):
    r"""Adaptive Top-label Calibration Error for binary tasks."""

    is_differentiable = False
    higher_is_better = False
    full_state_update = False

    confidences: list[Tensor]
    accuracies: list[Tensor]

    def __init__(
        self,
        n_bins: int = 10,
        norm: Literal["l1", "l2", "max"] = "l1",
        ignore_index: int | None = None,
        validate_args: bool = True,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        if ignore_index is not None:  # coverage: ignore
            raise ValueError("ignore_index is not supported for multiclass tasks.")

        if validate_args:
            _binary_calibration_error_arg_validation(n_bins, norm, ignore_index)
        self.n_bins = n_bins
        self.norm = norm

        self.add_state("confidences", [], dist_reduce_fx="cat")
        self.add_state("accuracies", [], dist_reduce_fx="cat")

    def update(self, probs: Tensor, targets: Tensor) -> None:
        """Update metric states with predictions and targets."""
        confidences, preds = torch.max(probs, 1 - probs), torch.round(probs)
        accuracies = preds == targets
        self.confidences.append(confidences.float())
        self.accuracies.append(accuracies.float())

    def compute(self) -> Tensor:
        """Compute metric."""
        confidences = dim_zero_cat(self.confidences)
        accuracies = dim_zero_cat(self.accuracies)
        return _ace_compute(confidences, accuracies, self.n_bins, norm=self.norm)


class MulticlassAdaptiveCalibrationError(Metric):
    r"""Adaptive Top-label Calibration Error for multiclass tasks."""

    is_differentiable: bool = False
    higher_is_better: bool = False
    full_state_update: bool = False

    confidences: list[Tensor]
    accuracies: list[Tensor]

    def __init__(
        self,
        num_classes: int,
        n_bins: int = 10,
        norm: Literal["l1", "l2", "max"] = "l1",
        ignore_index: int | None = None,
        validate_args: bool = True,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        if ignore_index is not None:  # coverage: ignore
            raise ValueError("ignore_index is not supported for multiclass tasks.")

        if validate_args:
            _multiclass_calibration_error_arg_validation(num_classes, n_bins, norm, ignore_index)
        self.n_bins = n_bins
        self.norm = norm

        self.add_state("confidences", [], dist_reduce_fx="cat")
        self.add_state("accuracies", [], dist_reduce_fx="cat")

    def update(self, probs: Tensor, targets: Tensor) -> None:
        """Update metric states with predictions and targets."""
        confidences, preds = torch.max(probs, 1)
        accuracies = preds == targets
        self.confidences.append(confidences.float())
        self.accuracies.append(accuracies.float())

    def compute(self) -> Tensor:
        """Compute metric."""
        confidences = dim_zero_cat(self.confidences)
        accuracies = dim_zero_cat(self.accuracies)
        return _ace_compute(confidences, accuracies, self.n_bins, norm=self.norm)


class AdaptiveCalibrationError:
    """`Adaptive Top-label Calibration Error`.

    Reference:
        Nixon et al. Measuring calibration in deep learning. In CVPRW, 2019.
    """

    def __new__(
        cls,
        task: Literal["binary", "multiclass"],
        num_bins: int = 10,
        norm: Literal["l1", "l2", "max"] = "l1",
        num_classes: int | None = None,
        ignore_index: int | None = None,
        validate_args: bool = True,
        **kwargs: Any,
    ) -> Metric:
        task = ClassificationTaskNoMultilabel.from_str(task)
        kwargs.update(
            {
                "n_bins": num_bins,
                "norm": norm,
                "ignore_index": ignore_index,
                "validate_args": validate_args,
            }
        )
        if task == ClassificationTaskNoMultilabel.BINARY:
            return BinaryAdaptiveCalibrationError(**kwargs)
        # task is ClassificationTaskNoMultilabel.MULTICLASS
        if not isinstance(num_classes, int):
            raise TypeError(
                f"`num_classes` is expected to be `int` but `{type(num_classes)} was passed.`"
            )
        return MulticlassAdaptiveCalibrationError(num_classes, **kwargs)



import torch
from torch import Tensor, nn
from torch.nn import functional as F


class DECLoss(nn.Module):
    def __init__(
        self,
        annealing_step: int | None = None,
        reg_weight: float | None = None,
        loss_type: str = "log",
        reduction: str | None = "mean",
    ) -> None:
        """The deep evidential classification loss.

        Args:
            annealing_step (int): Annealing step for the weight of the
            regularization term.
            reg_weight (float): Fixed weight of the regularization term.
            loss_type (str, optional): Specifies the loss type to apply to the
            Dirichlet parameters: ``'mse'`` | ``'log'`` | ``'digamma'``.
            reduction (str, optional): Specifies the reduction to apply to the
            output:``'none'`` | ``'mean'`` | ``'sum'``.

        Reference:
            Sensoy, M., Kaplan, L., & Kandemir, M. (2018). Evidential deep
            learning to quantify classification uncertainty. NeurIPS 2018.
            https://arxiv.org/abs/1806.01768.
        """
        super().__init__()

        if reg_weight is not None and (reg_weight < 0):
            raise ValueError(
                f"The regularization weight should be non-negative, but got {reg_weight}."
            )
        self.reg_weight = reg_weight

        if annealing_step is not None and (annealing_step <= 0):
            raise ValueError(f"The annealing step should be positive, but got {annealing_step}.")
        self.annealing_step = annealing_step

        if reduction not in ("none", "mean", "sum") and reduction is not None:
            raise ValueError(f"{reduction} is not a valid value for reduction.")
        self.reduction = reduction

        if loss_type not in ["mse", "log", "digamma"]:
            raise ValueError(f"{loss_type} is not a valid value for mse/log/digamma loss.")
        self.loss_type = loss_type

    def _mse_loss(self, evidence: Tensor, targets: Tensor) -> Tensor:
        evidence = torch.relu(evidence)
        alpha = evidence + 1.0
        strength = torch.sum(alpha, dim=1, keepdim=True)
        loglikelihood_err = torch.sum((targets - (alpha / strength)) ** 2, dim=1, keepdim=True)
        loglikelihood_var = torch.sum(
            alpha * (strength - alpha) / (strength * strength * (strength + 1)),
            dim=1,
            keepdim=True,
        )
        return loglikelihood_err + loglikelihood_var

    def _log_loss(self, evidence: Tensor, targets: Tensor) -> Tensor:
        evidence = torch.relu(evidence)
        alpha = evidence + 1.0
        strength = alpha.sum(dim=-1, keepdim=True)
        return torch.sum(
            targets * (torch.log(strength) - torch.log(alpha)),
            dim=1,
            keepdim=True,
        )

    def _digamma_loss(self, evidence: Tensor, targets: Tensor) -> Tensor:
        evidence = torch.relu(evidence)
        alpha = evidence + 1.0
        strength = alpha.sum(dim=-1, keepdim=True)
        return torch.sum(
            targets * (torch.digamma(strength) - torch.digamma(alpha)),
            dim=1,
            keepdim=True,
        )

    def _kldiv_reg(
        self,
        evidence: Tensor,
        targets: Tensor,
    ) -> Tensor:
        num_classes = evidence.size()[-1]
        evidence = torch.relu(evidence)
        alpha = evidence + 1.0

        kl_alpha = (alpha - 1) * (1 - targets) + 1

        ones = torch.ones([1, num_classes], dtype=evidence.dtype, device=evidence.device)
        sum_kl_alpha = torch.sum(kl_alpha, dim=1, keepdim=True)
        first_term = (
            torch.lgamma(sum_kl_alpha)
            - torch.lgamma(kl_alpha).sum(dim=1, keepdim=True)
            + torch.lgamma(ones).sum(dim=1, keepdim=True)
            - torch.lgamma(ones.sum(dim=1, keepdim=True))
        )
        second_term = torch.sum(
            (kl_alpha - ones) * (torch.digamma(kl_alpha) - torch.digamma(sum_kl_alpha)),
            dim=1,
            keepdim=True,
        )
        return first_term + second_term

    def forward(
        self,
        evidence: Tensor,
        targets: Tensor,
        current_epoch: int | None = None,
    ) -> Tensor:
        if self.annealing_step is not None and self.annealing_step > 0 and current_epoch is None:
            raise ValueError(
                "The epoch num should be positive when \
                annealing_step is settled, but got "
                f"{current_epoch}."
            )

        if targets.ndim != 1:  # if no mixup or cutmix
            raise NotImplementedError("DECLoss does not yet support mixup/cutmix.")
        # TODO: handle binary
        targets = F.one_hot(targets, num_classes=evidence.size()[-1])

        if self.loss_type == "mse":
            loss_dirichlet = self._mse_loss(evidence, targets)
        elif self.loss_type == "log":
            loss_dirichlet = self._log_loss(evidence, targets)
        else:  # self.loss_type == "digamma"
            loss_dirichlet = self._digamma_loss(evidence, targets)

        if self.reg_weight is None and self.annealing_step is None:
            annealing_coef = 0
        elif self.annealing_step is None and self.reg_weight > 0:
            annealing_coef = self.reg_weight
        else:
            annealing_coef = torch.min(
                input=torch.tensor(1.0, dtype=evidence.dtype),
                other=torch.tensor(current_epoch / self.annealing_step, dtype=evidence.dtype),
            )

        loss_reg = self._kldiv_reg(evidence, targets)
        loss = loss_dirichlet + annealing_coef * loss_reg
        if self.reduction == "mean":
            loss = loss.mean()
        elif self.reduction == "sum":
            loss = loss.sum()
        return loss


class ConfidencePenaltyLoss(nn.Module):
    def __init__(
        self,
        reg_weight: float = 1,
        reduction: str | None = "mean",
        eps: float = 1e-6,
    ) -> None:
        """The Confidence Penalty Loss.

        Args:
            reg_weight (float, optional): The weight of the regularization term.
            reduction (str, optional): specifies the reduction to apply to the
            output:``'none'`` | ``'mean'`` | ``'sum'``. Defaults to "mean".
            eps (float, optional): A small value to avoid numerical instability.
                Defaults to 1e-6.

        Reference:
            Gabriel Pereyra: Regularizing neural networks by penalizing
            confident output distributions. https://arxiv.org/pdf/1701.06548.

        """
        super().__init__()
        if reduction is None:
            reduction = "none"
        if reduction not in ("none", "mean", "sum"):
            raise ValueError(f"{reduction} is not a valid value for reduction.")
        self.reduction = reduction

        if eps < 0:
            raise ValueError(f"The epsilon value should be non-negative, but got {eps}.")
        self.eps = eps
        if reg_weight < 0:
            raise ValueError(
                f"The regularization weight should be non-negative, but got {reg_weight}."
            )
        self.reg_weight = reg_weight

    def forward(self, logits: Tensor, targets: Tensor) -> Tensor:
        """Compute the Confidence Penalty loss.

        Args:
            logits (Tensor): The inputs of the Bayesian Neural Network
            targets (Tensor): The target values

        Returns:
            Tensor: The Confidence Penalty loss
        """
        probs = F.softmax(logits, dim=1)
        ce_loss = F.cross_entropy(logits, targets, reduction=self.reduction)
        reg_loss = torch.log(torch.tensor(logits.shape[-1], device=probs.device)) + (
            probs * torch.log(probs + self.eps)
        ).sum(dim=-1)
        if self.reduction == "sum":
            return ce_loss + self.reg_weight * reg_loss.sum()
        if self.reduction == "mean":
            return ce_loss + self.reg_weight * reg_loss.mean()
        return ce_loss + self.reg_weight * reg_loss


class ConflictualLoss(nn.Module):
    def __init__(
        self,
        reg_weight: float = 1,
        reduction: str | None = "mean",
    ) -> None:
        r"""The Conflictual Loss.

        Args:
            reg_weight (float, optional): The weight of the regularization term.
            reduction (str, optional): specifies the reduction to apply to the
            output:``'none'`` | ``'mean'`` | ``'sum'``.

        Reference:
            `Mohammed Fellaji et al. On the Calibration of Epistemic Uncertainty:
            Principles, Paradoxes and Conflictual Loss <https://arxiv.org/pdf/2407.12211>`_.
        """
        super().__init__()
        if reduction is None:
            reduction = "none"
        if reduction not in ("none", "mean", "sum"):
            raise ValueError(f"{reduction} is not a valid value for reduction.")
        self.reduction = reduction
        if reg_weight < 0:
            raise ValueError(
                f"The regularization weight should be non-negative, but got {reg_weight}."
            )
        self.reg_weight = reg_weight

    def forward(self, logits: Tensor, targets: Tensor) -> Tensor:
        """Compute the conflictual loss.

        Args:
            logits (Tensor): The outputs of the model.
            targets (Tensor): The target values.

        Returns:
            Tensor: The conflictual loss.
        """
        class_index = torch.randint(
            0, logits.shape[-1], (1,), dtype=torch.long, device=logits.device
        )
        ce_loss = F.cross_entropy(logits, targets, reduction=self.reduction)
        reg_loss = -F.log_softmax(logits, dim=1)[:, class_index]
        if self.reduction == "sum":
            return ce_loss + self.reg_weight * reg_loss.sum()
        if self.reduction == "mean":
            return ce_loss + self.reg_weight * reg_loss.mean()
        return ce_loss + self.reg_weight * reg_loss


class FocalLoss(nn.Module):
    def __init__(
        self,
        gamma: float,
        alpha: Tensor | None = None,
        reduction: str = "mean",
    ) -> None:
        """Focal-Loss for classification tasks.

        Args:
            gamma (float, optional): A constant, as described in the paper.
            alpha (Tensor, optional): Weights for each class. Defaults to None.
            reduction (str, optional): 'mean', 'sum' or 'none'.
                Defaults to 'mean'.

        Reference:
            Lin, T.-Y., Goyal, P., Girshick, R., He, K., & Dollár, P. (2017).
            Focal Loss for Dense Object Detection. TPAMI 2020.

        Implementation:
            Inspired by github.com/AdeelH/pytorch-multi-class-focal-loss.
        """
        if reduction not in ("none", "mean", "sum") and reduction is not None:
            raise ValueError(f"{reduction} is not a valid value for reduction.")
        self.reduction = reduction

        if gamma < 0:
            raise ValueError(
                f"The gamma term of the focal loss should be non-negative, but got {gamma}."
            )
        self.gamma = gamma

        super().__init__()
        self.alpha = alpha
        self.nll_loss = nn.NLLLoss(weight=alpha, reduction="none")

    def forward(self, x: Tensor, y: Tensor) -> Tensor:
        log_p = F.log_softmax(x, dim=-1)
        ce = self.nll_loss(log_p, y)

        all_rows = torch.arange(len(x))
        log_pt = log_p[all_rows, y]

        pt = log_pt.exp()
        focal_term = (1 - pt) ** self.gamma

        loss = focal_term * ce

        if self.reduction == "mean":
            return loss.mean()
        if self.reduction == "sum":
            return loss.sum()
        return loss


class BCEWithLogitsLSLoss(nn.BCEWithLogitsLoss):
    def __init__(
        self,
        weight: Tensor | None = None,
        reduction: str = "mean",
        label_smoothing: float = 0.0,
    ) -> None:
        """Binary Cross Entropy with Logits Loss with label smoothing.

        The original PyTorch implementation of the BCEWithLogitsLoss does not
        support label smoothing. This implementation adds label smoothing to
        the BCEWithLogitsLoss.

        Args:
            weight (Tensor, optional): A manual rescaling weight given to the
                loss of each batch element. If given, has to be a Tensor of size
                "nbatch". Defaults to None.
            reduction (str, optional): Specifies the reduction to apply to the
                output: 'none' | 'mean' | 'sum'. 'none': no reduction will be applied,
                'mean': the sum of the output will be divided by the number of
                elements in the output, 'sum': the output will be summed. Defaults
                to 'mean'.
            label_smoothing (float, optional): The label smoothing factor. Defaults
                to 0.0.
        """
        super().__init__(weight=weight, reduction=reduction)
        self.label_smoothing = label_smoothing

    def forward(self, preds: Tensor, targets: Tensor) -> Tensor:
        if self.label_smoothing == 0.0:
            return super().forward(preds, targets.type_as(preds))
        targets = targets.float()
        targets = targets * (1 - self.label_smoothing) + self.label_smoothing / 2
        loss = targets * F.logsigmoid(preds) + (1 - targets) * F.logsigmoid(-preds)
        if self.weight is not None:
            loss = loss * self.weight
        if self.reduction == "mean":
            return -loss.mean()
        if self.reduction == "sum":
            return -loss.sum()
        return -loss



import torch
from torch import Tensor, nn
from torch.distributions import Independent

from torch_uncertainty.layers.bayesian import bayesian_modules
from torch_uncertainty.utils.distributions import get_dist_class


class KLDiv(nn.Module):
    def __init__(self, model: nn.Module) -> None:
        """KL divergence loss for Bayesian Neural Networks. Gathers the KL from the
        modules computed in the forward passes.

        Args:
            model (nn.Module): Bayesian Neural Network
        """
        super().__init__()
        self.model = model

    def forward(self) -> Tensor:
        return self._kl_div()

    def _kl_div(self) -> Tensor:
        """Gathers pre-computed KL-Divergences from :attr:`model`."""
        kl_divergence = torch.zeros(1)
        count = 0
        for module in self.model.modules():
            if isinstance(module, bayesian_modules):
                kl_divergence = kl_divergence.to(device=module.lvposterior.device)
                kl_divergence += module.lvposterior - module.lprior
                count += 1
        return kl_divergence / count


class ELBOLoss(nn.Module):
    def __init__(
        self,
        model: nn.Module | None,
        inner_loss: nn.Module,
        kl_weight: float,
        num_samples: int,
        dist_family: str | None = None,
    ) -> None:
        """The Evidence Lower Bound (ELBO) loss for Bayesian Neural Networks.

        ELBO loss for Bayesian Neural Networks. Use this loss function with the
        objective that you seek to minimize as :attr:`inner_loss`.

        Args:
            model (nn.Module): The Bayesian Neural Network to compute the loss for
            inner_loss (nn.Module): The loss function to use during training
            kl_weight (float): The weight of the KL divergence term
            num_samples (int): The number of samples to use for the ELBO loss
            dist_family (str, optional): The distribution family to use for the
                output of the model. ``None`` means point-wise prediction. Defaults to ``None``.

        Note:
            Set the model to None if you use the ELBOLoss within
            the ClassificationRoutine. It will get filled automatically.
        """
        super().__init__()
        _elbo_loss_checks(inner_loss, kl_weight, num_samples)
        self.set_model(model)

        self.inner_loss = inner_loss
        self.kl_weight = kl_weight
        self.num_samples = num_samples
        self.dist_family = dist_family

    def forward(self, inputs: Tensor, targets: Tensor) -> Tensor:
        """Gather the KL divergence from the Bayesian modules and aggregate
        the ELBO loss for a given network.

        Args:
            inputs (Tensor): The inputs of the Bayesian Neural Network
            targets (Tensor): The target values

        Returns:
            Tensor: The aggregated ELBO loss
        """
        aggregated_elbo = torch.zeros(1, device=inputs.device)
        dist_class = get_dist_class(self.dist_family) if self.dist_family is not None else None
        for _ in range(self.num_samples):
            out = self.model(inputs)
            if dist_class is not None:
                # Wrap the distribution in an Independent distribution for log_prob computation.
                out = Independent(dist_class(**out), 1)
            aggregated_elbo += self.inner_loss(out, targets)
            # TODO: This shouldn't be necessary
            aggregated_elbo += self.kl_weight * self._kl_div().to(inputs.device)
        return aggregated_elbo / self.num_samples

    def set_model(self, model: nn.Module | None) -> None:
        self.model = model
        if model is not None:
            self._kl_div = KLDiv(model)


def _elbo_loss_checks(inner_loss: nn.Module, kl_weight: float, num_samples: int) -> None:
    if isinstance(inner_loss, type):
        raise TypeError(f"The inner_loss should be an instance of a class.Got {inner_loss}.")

    if kl_weight < 0:
        raise ValueError(f"The KL weight should be non-negative. Got {kl_weight}.")

    if num_samples < 1:
        raise ValueError(f"The number of samples should not be lower than 1. Got {num_samples}.")
    if not isinstance(num_samples, int):
        raise TypeError(f"The number of samples should be an integer. Got {type(num_samples)}.")



from collections.abc import Callable
from typing import Literal

import torch
from einops import rearrange
from torch import nn
from torch.nn.functional import relu

from .std import _WideResNet

__all__ = [
    "mimo_wideresnet28x10",
]


class _MIMOWideResNet(_WideResNet):
    def __init__(
        self,
        depth: int,
        widen_factor: int,
        in_channels: int,
        num_classes: int,
        num_estimators: int,
        conv_bias: bool,
        dropout_rate: float,
        groups: int = 1,
        style: Literal["imagenet", "cifar"] = "imagenet",
        activation_fn: Callable = relu,
        normalization_layer: type[nn.Module] = nn.BatchNorm2d,
    ) -> None:
        super().__init__(
            depth,
            widen_factor=widen_factor,
            in_channels=in_channels * num_estimators,
            num_classes=num_classes * num_estimators,
            conv_bias=conv_bias,
            dropout_rate=dropout_rate,
            groups=groups,
            style=style,
            activation_fn=activation_fn,
            normalization_layer=normalization_layer,
        )
        self.num_estimators = num_estimators

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if not self.training:
            x = x.repeat(self.num_estimators, 1, 1, 1)
        out = rearrange(x, "(m b) c h w -> b (m c) h w", m=self.num_estimators)
        return rearrange(super().forward(out), "b (m d) -> (m b) d", m=self.num_estimators)


def mimo_wideresnet28x10(
    in_channels: int,
    num_classes: int,
    num_estimators: int,
    conv_bias: bool = True,
    dropout_rate: float = 0.3,
    groups: int = 1,
    style: Literal["imagenet", "cifar"] = "imagenet",
    activation_fn: Callable = relu,
    normalization_layer: type[nn.Module] = nn.BatchNorm2d,
) -> _MIMOWideResNet:
    return _MIMOWideResNet(
        depth=28,
        widen_factor=10,
        in_channels=in_channels,
        num_classes=num_classes,
        num_estimators=num_estimators,
        conv_bias=conv_bias,
        dropout_rate=dropout_rate,
        groups=groups,
        style=style,
        activation_fn=activation_fn,
        normalization_layer=normalization_layer,
    )



from pathlib import Path

import torch
from torch.nn.common_types import _size_2_t
from torch.nn.modules.utils import _pair
from torchvision import tv_tensors
from torchvision.datasets import VisionDataset
from torchvision.transforms import v2

from torch_uncertainty.datamodules import TUDataModule
from torch_uncertainty.transforms import RandomRescale
from torch_uncertainty.utils.misc import create_train_val_split


class DepthDataModule(TUDataModule):
    def __init__(
        self,
        dataset: type[VisionDataset],
        root: str | Path,
        batch_size: int,
        min_depth: float,
        max_depth: float,
        crop_size: _size_2_t,
        eval_size: _size_2_t,
        val_split: float | None = None,
        num_workers: int = 1,
        pin_memory: bool = True,
        persistent_workers: bool = True,
    ) -> None:
        r"""Base depth datamodule.

        Args:
            dataset (type[VisionDataset]): Dataset class to use.
            root (str or Path): Root directory of the datasets.
            batch_size (int): Number of samples per batch.
            min_depth (float, optional): Minimum depth value for evaluation.
            max_depth (float, optional): Maximum depth value for training and
                evaluation.
            crop_size (sequence or int, optional): Desired input image and
                depth mask sizes during training. If :attr:`crop_size` is an
                int instead of sequence like :math:`(H, W)`, a square crop
                :math:`(\text{size},\text{size})` is made. If provided a sequence
                of length :math:`1`, it will be interpreted as
                :math:`(\text{size[0]},\text{size[1]})`.
            eval_size (sequence or int, optional): Desired input image and
                depth mask sizes during evaluation. If size is an int,
                smaller edge of the images will be matched to this number, i.e.,
                :math:`\text{height}>\text{width}`, then image will be rescaled to
                :math:`(\text{size}\times\text{height}/\text{width},\text{size})`.
            val_split (float or None, optional): Share of training samples to use
                for validation.
            num_workers (int, optional): Number of dataloaders to use.
            pin_memory (bool, optional):  Whether to pin memory.
            persistent_workers (bool, optional): Whether to use persistent workers.
        """
        super().__init__(
            root=root,
            batch_size=batch_size,
            val_split=val_split,
            num_workers=num_workers,
            pin_memory=pin_memory,
            persistent_workers=persistent_workers,
        )

        self.dataset = dataset
        self.min_depth = min_depth
        self.max_depth = max_depth
        self.crop_size = _pair(crop_size)
        self.eval_size = _pair(eval_size)

        self.train_transform = v2.Compose(
            [
                RandomRescale(min_scale=0.5, max_scale=2.0),
                v2.RandomCrop(
                    size=self.crop_size,
                    pad_if_needed=True,
                    fill={tv_tensors.Image: 0, tv_tensors.Mask: float("nan")},
                ),
                v2.RandomHorizontalFlip(),
                v2.ToDtype(
                    dtype={
                        tv_tensors.Image: torch.float32,
                        "others": None,
                    },
                    scale=True,
                ),
                v2.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
            ]
        )
        self.test_transform = v2.Compose(
            [
                v2.Resize(size=self.eval_size),
                v2.ToDtype(
                    dtype={
                        tv_tensors.Image: torch.float32,
                        "others": None,
                    },
                    scale=True,
                ),
                v2.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
            ]
        )

    def prepare_data(self) -> None:  # coverage: ignore
        self.dataset(
            root=self.root,
            split="train",
            max_depth=self.max_depth,
            download=True,
        )
        self.dataset(root=self.root, split="val", max_depth=self.max_depth, download=True)

    def setup(self, stage: str | None = None) -> None:
        if stage == "fit" or stage is None:
            full = self.dataset(
                root=self.root,
                max_depth=self.max_depth,
                split="train",
                transforms=self.train_transform,
            )

            if self.val_split is not None:
                self.train, self.val = create_train_val_split(
                    full,
                    self.val_split,
                    self.test_transform,
                )
                self.val.min_depth = self.min_depth
            else:
                self.train = full
                self.val = self.dataset(
                    root=self.root,
                    min_depth=self.min_depth,
                    max_depth=self.max_depth,
                    split="val",
                    transforms=self.test_transform,
                )

        if stage == "test" or stage is None:
            self.test = self.dataset(
                root=self.root,
                min_depth=self.min_depth,
                max_depth=self.max_depth,
                split="val",
                transforms=self.test_transform,
            )

        if stage not in ["fit", "test", None]:
            raise ValueError(f"Stage {stage} is not supported.")



import torch

from torch_uncertainty.models.segmentation.segformer import (
    seg_former,
)


class TestSegformer:
    """Testing the Segformer class."""

    @torch.no_grad()
    def test_main(self):
        model = seg_former(10, 0)
        seg_former(10, 1)
        model(torch.randn(1, 3, 32, 32))



import torch
from torch import Tensor, nn


class _FilterResponseNormNd(nn.Module):
    def __init__(
        self,
        dimension: int,
        num_channels: int,
        eps: float = 1e-6,
        device=None,
        dtype=None,
    ) -> None:
        """N-dimensional Filter Response Normalization layer.

        Args:
            dimension (int): Dimension of the input tensor.
            num_channels (int): Number of channels.
            eps (float, optional): Epsilon. Defaults to 1e-6.
            device (optional): Device. Defaults to None.
            dtype (optional): Data type. Defaults to None.
        """
        super().__init__()
        if dimension < 1 or not isinstance(dimension, int):
            raise ValueError(
                f"dimension should be an integer greater or equal than 1. Got {dimension}."
            )
        self.dimension = dimension

        if num_channels < 1 or not isinstance(num_channels, int):
            raise ValueError(
                f"num_channels should be an integer greater or equal than 1. Got {num_channels}."
            )
        shape = (1, num_channels) + (1,) * dimension
        self.eps = eps

        self.tau = nn.Parameter(torch.zeros(shape, device=device, dtype=dtype))
        self.beta = nn.Parameter(torch.zeros(shape, device=device, dtype=dtype))
        self.gamma = nn.Parameter(torch.ones(shape, device=device, dtype=dtype))

    def forward(self, x: Tensor) -> Tensor:
        nu2 = torch.mean(x**2, dim=list(range(-self.dimension, 0)), keepdim=True)
        x = x * torch.rsqrt(nu2 + self.eps)
        y = self.gamma * x + self.beta
        return torch.max(y, self.tau)


class FilterResponseNorm1d(_FilterResponseNormNd):
    def __init__(self, num_channels: int, eps: float = 1e-6, device=None, dtype=None) -> None:
        """1-dimensional Filter Response Normalization layer.

        Args:
            num_channels (int): Number of channels.
            eps (float, optional): Epsilon. Defaults to 1e-6.
            device (optional): Device. Defaults to None.
            dtype (optional): Data type. Defaults to None.
        """
        super().__init__(
            dimension=1,
            num_channels=num_channels,
            eps=eps,
            device=device,
            dtype=dtype,
        )


class FilterResponseNorm2d(_FilterResponseNormNd):
    def __init__(self, num_channels: int, eps: float = 1e-6, device=None, dtype=None) -> None:
        """2-dimensional Filter Response Normalization layer.

        Args:
            num_channels (int): Number of channels.
            eps (float, optional): Epsilon. Defaults to 1e-6.
            device (optional): Device. Defaults to None.
            dtype (optional): Data type. Defaults to None.
        """
        super().__init__(
            dimension=2,
            num_channels=num_channels,
            eps=eps,
            device=device,
            dtype=dtype,
        )


class FilterResponseNorm3d(_FilterResponseNormNd):
    def __init__(self, num_channels: int, eps: float = 1e-6, device=None, dtype=None) -> None:
        """3-dimensional Filter Response Normalization layer.

        Args:
            num_channels (int): Number of channels.
            eps (float, optional): Epsilon. Defaults to 1e-6.
            device (optional): Device. Defaults to None.
            dtype (optional): Data type. Defaults to None.
        """
        super().__init__(
            dimension=3,
            num_channels=num_channels,
            eps=eps,
            device=device,
            dtype=dtype,
        )



from typing import Literal

from torch import nn
from torch.optim import Optimizer

from torch_uncertainty.models import mc_dropout
from torch_uncertainty.models.wideresnet import (
    batched_wideresnet28x10,
    masked_wideresnet28x10,
    mimo_wideresnet28x10,
    packed_wideresnet28x10,
    wideresnet28x10,
)
from torch_uncertainty.routines.classification import (
    ClassificationRoutine,
)
from torch_uncertainty.transforms import MIMOBatchFormat, RepeatTarget

ENSEMBLE_METHODS = ["packed", "batched", "masked", "mimo", "mc-dropout"]


class WideResNetBaseline(ClassificationRoutine):
    versions = {
        "std": [wideresnet28x10],
        "mc-dropout": [wideresnet28x10],
        "packed": [packed_wideresnet28x10],
        "batched": [batched_wideresnet28x10],
        "masked": [masked_wideresnet28x10],
        "mimo": [mimo_wideresnet28x10],
    }

    def __init__(
        self,
        num_classes: int,
        in_channels: int,
        loss: nn.Module,
        version: Literal["std", "mc-dropout", "packed", "batched", "masked", "mimo"],
        style: str = "imagenet",
        num_estimators: int = 1,
        dropout_rate: float = 0.0,
        optim_recipe: dict | Optimizer | None = None,
        mixup_params: dict | None = None,
        groups: int = 1,
        last_layer_dropout: bool = False,
        scale: float | None = None,
        alpha: int | None = None,
        gamma: int = 1,
        rho: float = 1.0,
        batch_repeat: int = 1,
        ood_criterion: Literal["msp", "logit", "energy", "entropy", "mi", "vr"] = "msp",
        log_plots: bool = False,
        save_in_csv: bool = False,
        calibration_set: Literal["val", "test"] = "val",
        eval_ood: bool = False,
        eval_shift: bool = False,
        eval_grouping_loss: bool = False,
    ) -> None:
        r"""Wide-ResNet28x10 backbone baseline for classification providing support
        for various versions.

        Args:
            num_classes (int): Number of classes to predict.
            in_channels (int): Number of input channels.
            loss (nn.Module): Training loss.
            optim_recipe (Any): optimization recipe, corresponds to
                what expect the `LightningModule.configure_optimizers()
                <https://pytorch-lightning.readthedocs.io/en/stable/common/lightning_module.html#configure-optimizers>`_
                method.
            version (str):
                Determines which Wide-ResNet version to use:

                - ``"std"``: original Wide-ResNet
                - ``"mc-dropout"``: Monte Carlo Dropout Wide-ResNet
                - ``"packed"``: Packed-Ensembles Wide-ResNet
                - ``"batched"``: BatchEnsemble Wide-ResNet
                - ``"masked"``: Masksemble Wide-ResNet
                - ``"mimo"``: MIMO Wide-ResNet

            style (bool, optional): (str, optional): Which ResNet style to use.
            Defaults to ``imagenet``.
            num_estimators (int, optional): Number of estimators in the ensemble.
                Only used if :attr:`version` is either ``"packed"``, ``"batched"``
                or ``"masked"`` Defaults to ``None``.
            dropout_rate (float, optional): Dropout rate. Defaults to ``0.0``.
            mixup_params (dict, optional): Mixup parameters. Can include mixtype,
                mixmode, dist_sim, kernel_tau_max, kernel_tau_std,
                mixup_alpha, and cutmix_alpha. If None, no augmentations.
                Defaults to ``None``.
            last_layer_dropout (bool): whether to apply dropout to the last layer only.
            groups (int, optional): Number of groups in convolutions. Defaults to
                ``1``.
            scale (float, optional): Expansion factor affecting the width of the
                estimators. Only used if :attr:`version` is ``"masked"``. Defaults
                to ``None``.
            alpha (float, optional): Expansion factor affecting the width of the
                estimators. Only used if :attr:`version` is ``"packed"``. Defaults
                to ``None``.
            gamma (int, optional): Number of groups within each estimator. Only
                used if :attr:`version` is ``"packed"`` and scales with
                :attr:`groups`. Defaults to ``1s``.
            rho (float, optional): Probability that all estimators share the same
                input. Only used if :attr:`version` is ``"mimo"``. Defaults to
                ``1``.
            batch_repeat (int, optional): Number of times to repeat the batch. Only
                used if :attr:`version` is ``"mimo"``. Defaults to ``1``.
            ood_criterion (str, optional): OOD criterion. Defaults to ``"msp"``.
                MSP is the maximum softmax probability, logit is the maximum
                logit, entropy is the entropy of the mean prediction, mi is the
                mutual information of the ensemble and vr is the variation ratio
                of the ensemble.
            log_plots (bool, optional): Indicates whether to log the plots or not.
                Defaults to ``False``.
            save_in_csv (bool, optional): Indicates whether to save the results in
                a csv file or not. Defaults to ``False``.
            calibration_set (Callable, optional): Calibration set. Defaults to
                ``None``.
            eval_ood (bool, optional): Indicates whether to evaluate the
                OOD detection or not. Defaults to ``False``.
            eval_shift (bool): Whether to evaluate on shifted data. Defaults to
                ``False``.
            eval_grouping_loss (bool, optional): Indicates whether to evaluate the
                grouping loss or not. Defaults to ``False``.

        Raises:
            ValueError: If :attr:`version` is not either ``"std"``,
                ``"packed"``, ``"batched"`` or ``"masked"``.

        Returns:
            LightningModule: Wide-ResNet baseline ready for training and
                evaluation.
        """
        params = {
            "conv_bias": False,
            "dropout_rate": dropout_rate,
            "groups": groups,
            "in_channels": in_channels,
            "num_classes": num_classes,
            "style": style,
        }

        format_batch_fn = nn.Identity()

        if version not in self.versions:
            raise ValueError(f"Unknown version: {version}")

        if version in ENSEMBLE_METHODS:
            params |= {
                "num_estimators": num_estimators,
            }

            if version != "mc-dropout":
                format_batch_fn = RepeatTarget(num_repeats=num_estimators)

        if version == "packed":
            params |= {
                "alpha": alpha,
                "gamma": gamma,
            }

        elif version == "masked":
            params |= {
                "scale": scale,
            }

        elif version == "mimo":
            format_batch_fn = MIMOBatchFormat(
                num_estimators=num_estimators,
                rho=rho,
                batch_repeat=batch_repeat,
            )

        if version == "mc-dropout":  # std wideRn don't have `num_estimators`
            del params["num_estimators"]

        model = self.versions[version][0](**params)

        if version == "mc-dropout":
            model = mc_dropout(
                model=model,
                num_estimators=num_estimators,
                last_layer=last_layer_dropout,
            )

        super().__init__(
            num_classes=num_classes,
            model=model,
            loss=loss,
            is_ensemble=version in ENSEMBLE_METHODS,
            format_batch_fn=format_batch_fn,
            optim_recipe=optim_recipe,
            mixup_params=mixup_params,
            eval_ood=eval_ood,
            eval_shift=eval_shift,
            eval_grouping_loss=eval_grouping_loss,
            ood_criterion=ood_criterion,
            log_plots=log_plots,
            save_in_csv=save_in_csv,
            calibration_set=calibration_set,
        )
        self.save_hyperparameters(ignore=["loss"])



from pathlib import Path
from typing import Literal

import torchvision.transforms as T
from torch import nn
from torch.utils.data import DataLoader
from torchvision.datasets import MNIST, FashionMNIST

from torch_uncertainty.datamodules import TUDataModule
from torch_uncertainty.datasets.classification import MNISTC, NotMNIST
from torch_uncertainty.transforms import Cutout
from torch_uncertainty.utils import create_train_val_split


class MNISTDataModule(TUDataModule):
    num_classes = 10
    num_channels = 1
    input_shape = (1, 28, 28)
    training_task = "classification"
    ood_datasets = ["fashion", "notMNIST"]
    mean = (0.1307,)
    std = (0.3081,)

    def __init__(
        self,
        root: str | Path,
        batch_size: int,
        eval_ood: bool = False,
        eval_shift: bool = False,
        ood_ds: Literal["fashion", "notMNIST"] = "fashion",
        val_split: float | None = None,
        num_workers: int = 1,
        basic_augment: bool = True,
        cutout: int | None = None,
        pin_memory: bool = True,
        persistent_workers: bool = True,
    ) -> None:
        """DataModule for MNIST.

        Args:
            root (str): Root directory of the datasets.
            eval_ood (bool): Whether to evaluate on out-of-distribution data.
                Defaults to ``False``.
            eval_shift (bool): Whether to evaluate on shifted data. Defaults to
                ``False``.
            batch_size (int): Number of samples per batch.
            ood_ds (str): Which out-of-distribution dataset to use. Defaults to
                ``"fashion"``; `fashion` stands for FashionMNIST and `notMNIST` for
                notMNIST.
            val_split (float): Share of samples to use for validation. Defaults
                to ``0.0``.
            num_workers (int): Number of workers to use for data loading. Defaults
                to ``1``.
            basic_augment (bool): Whether to apply base augmentations. Defaults to
                ``True``.
            cutout (int): Size of cutout to apply to images. Defaults to ``None``.
            pin_memory (bool): Whether to pin memory. Defaults to ``True``.
            persistent_workers (bool): Whether to use persistent workers. Defaults
                to ``True``.
        """
        super().__init__(
            root=root,
            batch_size=batch_size,
            val_split=val_split,
            num_workers=num_workers,
            pin_memory=pin_memory,
            persistent_workers=persistent_workers,
        )

        self.eval_ood = eval_ood
        self.eval_shift = eval_shift
        self.batch_size = batch_size

        self.dataset = MNIST

        if ood_ds == "fashion":
            self.ood_dataset = FashionMNIST
        elif ood_ds == "notMNIST":
            self.ood_dataset = NotMNIST
        else:
            raise ValueError(f"`ood_ds` should be in {self.ood_datasets}. Got {ood_ds}.")
        self.shift_dataset = MNISTC
        self.shift_severity = 1

        basic_transform = T.RandomCrop(28, padding=4) if basic_augment else nn.Identity()

        main_transform = Cutout(cutout) if cutout else nn.Identity()

        self.train_transform = T.Compose(
            [
                T.ToTensor(),
                basic_transform,
                main_transform,
                T.Normalize(mean=self.mean, std=self.std),
            ]
        )
        self.test_transform = T.Compose(
            [
                T.ToTensor(),
                T.CenterCrop(28),
                T.Normalize(mean=self.mean, std=self.std),
            ]
        )
        if self.eval_ood:  # NotMNIST has 3 channels
            self.ood_transform = T.Compose(
                [
                    T.ToTensor(),
                    T.Grayscale(num_output_channels=1),
                    T.CenterCrop(28),
                    T.Normalize(mean=self.mean, std=self.std),
                ]
            )

    def prepare_data(self) -> None:  # coverage: ignore
        """Download the datasets."""
        self.dataset(self.root, train=True, download=True)
        self.dataset(self.root, train=False, download=True)

        if self.eval_ood:
            self.ood_dataset(self.root, download=True)
        if self.eval_shift:
            self.shift_dataset(self.root, download=True)

    def setup(self, stage: Literal["fit", "test"] | None = None) -> None:
        if stage == "fit" or stage is None:
            full = self.dataset(
                self.root,
                train=True,
                download=False,
                transform=self.train_transform,
            )
            if self.val_split:
                self.train, self.val = create_train_val_split(
                    full,
                    self.val_split,
                    self.test_transform,
                )
            else:
                self.train = full
                self.val = self.dataset(
                    self.root,
                    train=False,
                    download=False,
                    transform=self.test_transform,
                )
        if stage == "test" or stage is None:
            self.test = self.dataset(
                self.root,
                train=False,
                download=False,
                transform=self.test_transform,
            )
        if stage not in ["fit", "test", None]:
            raise ValueError(f"Stage {stage} is not supported.")

        if self.eval_ood:
            self.ood = self.ood_dataset(
                self.root,
                download=False,
                transform=self.ood_transform,
            )
        if self.eval_shift:
            self.shift = self.shift_dataset(
                self.root,
                download=False,
                transform=self.test_transform,
            )

    def test_dataloader(self) -> list[DataLoader]:
        r"""Get the test dataloaders for MNIST.

        Return:
            list[DataLoader]: Dataloaders of the MNIST test set (in
                distribution data), FashionMNIST or NotMNIST test split
                (out-of-distribution data), and/or MNISTC (shifted data).
        """
        dataloader = [self._data_loader(self.test)]
        if self.eval_ood:
            dataloader.append(self._data_loader(self.ood))
        if self.eval_shift:
            dataloader.append(self._data_loader(self.shift))
        return dataloader



# ruff: noqa: F401
from .bank_marketing import BankMarketingDataModule
from .dota2_games import DOTA2GamesDataModule
from .htru2 import HTRU2DataModule
from .online_shoppers import OnlineShoppersDataModule
from .spam_base import SpamBaseDataModule
from .uci_classification import UCIClassificationDataModule



from functools import partial
from pathlib import Path

from torch import Generator
from torch.utils.data import random_split

from torch_uncertainty.datasets.regression import UCIRegression

from .abstract import TUDataModule


class UCIRegressionDataModule(TUDataModule):
    training_task = "regression"

    def __init__(
        self,
        root: str | Path,
        batch_size: int,
        dataset_name: str,
        val_split: float = 0.0,
        num_workers: int = 1,
        pin_memory: bool = True,
        persistent_workers: bool = True,
        input_shape: tuple[int, ...] | None = None,
        split_seed: int = 42,
    ) -> None:
        """The UCI regression datasets.

        Args:
            root (string): Root directory of the datasets.
            batch_size (int): The batch size for training and testing.
            dataset_name (string, optional): The name of the dataset. One of
                "boston-housing", "concrete", "energy", "kin8nm",
                "naval-propulsion-plant", "power-plant", "protein",
                "wine-quality-red", and "yacht".
            val_split (float, optional): Share of validation samples. Defaults
                to ``0``.
            num_workers (int, optional): How many subprocesses to use for data
                loading. Defaults to ``1``.
            pin_memory (bool, optional): Whether to pin memory in the GPU. Defaults
                to ``True``.
            persistent_workers (bool, optional): Whether to use persistent workers.
                Defaults to ``True``.
            input_shape (tuple, optional): The shape of the input data. Defaults to
                ``None``.
            split_seed (int, optional): The seed to use for splitting the dataset.
                Defaults to ``42``.
        """
        super().__init__(
            root=root,
            batch_size=batch_size,
            val_split=val_split,
            num_workers=num_workers,
            pin_memory=pin_memory,
            persistent_workers=persistent_workers,
        )

        self.dataset = partial(UCIRegression, dataset_name=dataset_name, seed=split_seed)
        self.input_shape = input_shape
        self.gen = Generator().manual_seed(split_seed)

    def prepare_data(self) -> None:
        """Download the dataset."""
        self.dataset(root=self.root, download=True)

    # ruff: noqa: ARG002
    def setup(self, stage: str | None = None) -> None:
        """Split the datasets into train, val, and test."""
        full = self.dataset(
            self.root,
            download=False,
        )
        self.train, self.test, self.val = random_split(
            full,
            [
                0.8 - self.val_split,
                0.2,
                self.val_split,
            ],
            generator=self.gen,
        )
        if self.val_split == 0:
            self.val = self.test



import pytest
import torch

from tests._dummies import dummy_model
from torch_uncertainty.models import deep_ensembles


class TestDeepEnsemblesModel:
    """Testing the deep_ensembles function."""

    def test_main(self):
        model_1 = dummy_model(1, 10)
        model_2 = dummy_model(1, 10)

        de = deep_ensembles([model_1, model_2])
        # Check B N C
        assert de(torch.randn(3, 4, 4)).shape == (6, 10)

    def test_list_and_num_estimators(self):
        model_1 = dummy_model(1, 10)
        model_2 = dummy_model(1, 10)
        with pytest.raises(ValueError):
            deep_ensembles([model_1, model_2], num_estimators=2)

    def test_list_singleton(self):
        model_1 = dummy_model(1, 10)

        deep_ensembles([model_1], num_estimators=2, reset_model_parameters=True)
        deep_ensembles(model_1, num_estimators=2, reset_model_parameters=False)

        with pytest.raises(ValueError):
            deep_ensembles([model_1], num_estimators=1)

    def test_error_prob_regression(self):
        # The output dicts will have different keys
        model_1 = dummy_model(1, 2, dist_family="normal")
        model_2 = dummy_model(1, 2, dist_family="nig")

        de = deep_ensembles([model_1, model_2], task="regression", probabilistic=True)

        with pytest.raises(ValueError):
            de(torch.randn(5, 1))

    def test_errors(self):
        model_1 = dummy_model(1, 10)
        with pytest.raises(ValueError):
            deep_ensembles(model_1, num_estimators=None)

        with pytest.raises(ValueError):
            deep_ensembles(model_1, num_estimators=-1)

        with pytest.raises(ValueError):
            deep_ensembles(model_1, num_estimators=1)

        with pytest.raises(ValueError):
            deep_ensembles(model_1, num_estimators=2, task="regression")

        with pytest.raises(ValueError):
            deep_ensembles(model_1, num_estimators=2, task="other")



import copy
import csv
from collections.abc import Callable
from pathlib import Path

import matplotlib.pyplot as plt
import torch
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from torch.utils.data import Dataset, random_split


def csv_writer(path: Path, dic: dict) -> None:
    """Write a dictionary to a csv file.

    Args:
        path (Path): Path to the csv file.
        dic (dict): Dictionary to write.
    """
    if path.is_file():  # Check that the file already exists
        append_mode = True
        rw_mode = "a"
    else:
        append_mode = False
        rw_mode = "w"
    # Write dic
    with path.open(rw_mode) as csvfile:
        writer = csv.writer(csvfile, delimiter=",")
        # Do not write header in append mode
        if append_mode is False:
            writer.writerow(dic.keys())
        writer.writerow([f"{elem:.4f}" for elem in dic.values()])


def plot_hist(
    conf: list[torch.Tensor],
    bins: int = 20,
    title: str = "Histogram with 'auto' bins",
    dpi: int = 60,
) -> tuple[Figure, Axes]:
    """Plot a confidence histogram.

    Args:
        conf (Any): The confidence values.
        bins (int, optional): The number of bins. Defaults to 20.
        title (str, optional): The title of the plot. Defaults to "Histogram
            with 'auto' bins".
        dpi (int, optional): The dpi of the plot. Defaults to 60.

    Returns:
        Tuple[Figure, Axes]: The figure and axes of the plot.
    """
    plt.rc("axes", axisbelow=True)
    fig, ax = plt.subplots(1, figsize=(7, 5), dpi=dpi)
    for i in [1, 0]:
        ax.hist(
            conf[i],
            bins=bins,
            density=True,
            label=["In-distribution", "Out-of-Distribution"][i],
            alpha=0.4,
            linewidth=1,
            edgecolor=["#0d559f", "#d45f00"][i],
            color=["#1f77b4", "#ff7f0e"][i],
        )

    ax.set_title(title)
    plt.grid(True, linestyle="--", alpha=0.7, zorder=0)
    plt.legend()
    fig.tight_layout()
    return fig, ax


def create_train_val_split(
    dataset: Dataset,
    val_split_rate: float,
    val_transforms: Callable | None = None,
) -> tuple[Dataset, Dataset]:
    train, val = random_split(dataset, [1 - val_split_rate, val_split_rate])
    val = copy.deepcopy(val)
    val.dataset.transform = val_transforms
    return train, val



import math
from typing import Literal

import torch
import torchvision.models as tv_models
from torch import Tensor, nn
from torch.nn import functional as F
from torchvision.models.densenet import DenseNet121_Weights, DenseNet161_Weights
from torchvision.models.resnet import (
    ResNet50_Weights,
    ResNet101_Weights,
    ResNeXt50_32X4D_Weights,
    ResNeXt101_32X8D_Weights,
)

from torch_uncertainty.layers.distributions import get_dist_conv_layer
from torch_uncertainty.models.utils import Backbone

resnet_feat_out_channels = [64, 256, 512, 1024, 2048]
resnet_feat_names = ["relu", "layer1", "layer2", "layer3", "layer4"]
densenet_feat_names = [
    "relu0",
    "pool0",
    "transition1",
    "transition2",
    "norm5",
]

bts_backbones = [
    "densenet121",
    "densenet161",
    "resnet50",
    "resnet101",
    "resnext50",
    "resnext101",
]


class AtrousBlock2d(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        dilation: int,
        norm_first: bool = True,
        norm_momentum: float = 0.1,
        **factory_kwargs,
    ):
        """Atrous block with 1x1 and 3x3 convolutions.

        Args:
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
            dilation (int): Dilation rate for the 3x3 convolution.
            norm_first (bool): Whether to apply normalization before the 1x1 convolution.
                Defaults to True.
            norm_momentum (float): Momentum for the normalization layer. Defaults to 0.1.
            factory_kwargs: Additional arguments for the PyTorch layers.
        """
        super().__init__()

        self.norm_first = norm_first
        if norm_first:
            self.first_norm = nn.BatchNorm2d(in_channels, momentum=norm_momentum, **factory_kwargs)

        self.conv1 = nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels * 2,
            bias=False,
            kernel_size=1,
            stride=1,
            padding=0,
            **factory_kwargs,
        )
        self.norm = nn.BatchNorm2d(out_channels * 2, momentum=norm_momentum, **factory_kwargs)
        self.conv2 = nn.Conv2d(
            in_channels=out_channels * 2,
            out_channels=out_channels,
            bias=False,
            kernel_size=3,
            stride=1,
            padding=(dilation, dilation),
            dilation=dilation,
            **factory_kwargs,
        )

    def forward(self, x: Tensor) -> Tensor:
        if self.norm_first:
            x = self.first_norm(x)
        out = F.relu(self.conv1(x))
        out = F.relu(self.norm(out))
        return self.conv2(out)


class UpConv2d(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        ratio: int = 2,
        **factory_kwargs,
    ):
        """Upsampling convolution.

        Args:
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
            ratio (int): Upsampling ratio.
            factory_kwargs: Additional arguments for the convolution layer.
        """
        super().__init__()
        self.conv = nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            bias=False,
            kernel_size=3,
            stride=1,
            padding=1,
            **factory_kwargs,
        )
        self.ratio = ratio

    def forward(self, x: Tensor) -> Tensor:
        out = F.interpolate(x, scale_factor=self.ratio, mode="nearest")
        return F.elu(self.conv(out))


class Reduction1x1(nn.Module):
    def __init__(
        self,
        num_in_filters: int,
        num_out_filters: int,
        max_depth: float,
        is_final: bool = False,
        **factory_kwargs,
    ):
        super().__init__()
        self.max_depth = max_depth
        self.is_final = is_final
        self.reduc = torch.nn.Sequential()

        while num_out_filters >= 4:
            if num_out_filters < 8:
                if self.is_final:
                    self.reduc.add_module(
                        "final",
                        torch.nn.Sequential(
                            nn.Conv2d(
                                num_in_filters,
                                out_channels=1,
                                bias=False,
                                kernel_size=1,
                                stride=1,
                                padding=0,
                                **factory_kwargs,
                            ),
                            nn.Sigmoid(),
                        ),
                    )
                else:
                    self.reduc.add_module(
                        "plane_params",
                        torch.nn.Conv2d(
                            num_in_filters,
                            out_channels=3,
                            bias=False,
                            kernel_size=1,
                            stride=1,
                            padding=0,
                            **factory_kwargs,
                        ),
                    )
                break

            self.reduc.add_module(
                f"inter_{num_in_filters}_{num_out_filters}",
                torch.nn.Sequential(
                    nn.Conv2d(
                        in_channels=num_in_filters,
                        out_channels=num_out_filters,
                        bias=False,
                        kernel_size=1,
                        stride=1,
                        padding=0,
                        **factory_kwargs,
                    ),
                    nn.ELU(),
                ),
            )

            num_in_filters = num_out_filters
            num_out_filters = num_out_filters // 2

    def forward(self, x: Tensor) -> Tensor:
        x = self.reduc.forward(x)
        if not self.is_final:
            theta = F.sigmoid(x[:, 0, :, :]) * math.pi / 3
            phi = F.sigmoid(x[:, 1, :, :]) * math.pi * 2
            dist = F.sigmoid(x[:, 2, :, :]) * self.max_depth
            x = torch.cat(
                [
                    torch.mul(torch.sin(theta), torch.cos(phi)).unsqueeze(1),
                    torch.mul(torch.sin(theta), torch.sin(phi)).unsqueeze(1),
                    torch.cos(theta).unsqueeze(1),
                    dist.unsqueeze(1),
                ],
                dim=1,
            )
        return x


class LocalPlanarGuidance(nn.Module):
    def __init__(self, up_ratio: int) -> None:
        super().__init__()
        self.register_buffer("u", torch.arange(up_ratio).reshape([1, 1, up_ratio]))
        self.register_buffer("v", torch.arange(up_ratio).reshape([1, up_ratio, 1]))
        self.up_ratio = up_ratio

    def forward(self, x: Tensor) -> Tensor:
        x_expanded = torch.repeat_interleave(
            torch.repeat_interleave(x, self.up_ratio, 2), self.up_ratio, 3
        )

        u = self.u.repeat(
            x.size(0),
            x.size(2) * self.up_ratio,
            x.size(3),
        )
        u = (u - (self.up_ratio - 1) * 0.5) / self.up_ratio

        v = self.v.repeat(
            x.size(0),
            x.size(2),
            x.size(3) * self.up_ratio,
        )
        v = (v - (self.up_ratio - 1) * 0.5) / self.up_ratio

        return x_expanded[:, 3, :, :] / (
            x_expanded[:, 0, :, :] * u + x_expanded[:, 1, :, :] * v + x_expanded[:, 2, :, :]
        )


class BTSBackbone(Backbone):
    def __init__(self, backbone_name: str, pretrained: bool) -> None:
        """BTS backbone.

        Args:
            backbone_name (str): Name of the backbone.
            pretrained (bool): Use a pretrained backbone.
        """
        if backbone_name == "densenet121":
            model = tv_models.densenet121(
                weights=DenseNet121_Weights.DEFAULT if pretrained else None
            ).features
            feat_names = densenet_feat_names
            self.feat_out_channels = [64, 64, 128, 256, 1024]
        elif backbone_name == "densenet161":
            model = tv_models.densenet161(
                weights=DenseNet161_Weights.DEFAULT if pretrained else None
            ).features
            feat_names = densenet_feat_names
            self.feat_out_channels = [96, 96, 192, 384, 2208]
        elif backbone_name == "resnet50":
            model = tv_models.resnet50(
                weights=ResNet50_Weights.IMAGENET1K_V2 if pretrained else None
            )
        elif backbone_name == "resnet101":
            model = tv_models.resnet101(
                weights=ResNet101_Weights.IMAGENET1K_V2 if pretrained else None
            )
        elif backbone_name == "resnext50":
            model = tv_models.resnext50_32x4d(
                weights=ResNeXt50_32X4D_Weights.IMAGENET1K_V2 if pretrained else None
            )
        else:  # backbone_name == "resnext101":
            model = tv_models.resnext101_32x8d(
                weights=ResNeXt101_32X8D_Weights.IMAGENET1K_V2 if pretrained else None
            )
        if "res" in backbone_name:  # remove classification heads from ResNets
            feat_names = resnet_feat_names
            self.feat_out_channels = resnet_feat_out_channels
            model.avgpool = nn.Identity()
            model.fc = nn.Identity()
        super().__init__(model=model, feat_names=feat_names)


class BTSDecoder(nn.Module):
    """BTS decoder.

    Args:
        max_depth (float): The maximum predicted depth.
        feat_out_channels (list[int]): The number of output channels from the backbone.
        num_features (int): The number of features to use in the decoder.
        dist_family (str | None, optional): The distribution family name. ``None`` means point-wise
            prediction. Defaults to ``None``.
    """

    def __init__(
        self,
        max_depth: float,
        feat_out_channels: list[int],
        num_features: int,
        dist_family: str | None = None,
    ):
        super().__init__()
        self.max_depth = max_depth

        self.upconv5 = UpConv2d(in_channels=feat_out_channels[4], out_channels=num_features)
        self.bn5 = nn.BatchNorm2d(num_features, momentum=0.01, affine=True)

        self.conv5 = nn.Conv2d(
            num_features + feat_out_channels[3],
            num_features,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=False,
        )
        self.upconv4 = UpConv2d(in_channels=num_features, out_channels=num_features // 2)
        self.bn4 = nn.BatchNorm2d(num_features // 2, momentum=0.01, affine=True)
        self.conv4 = nn.Conv2d(
            num_features // 2 + feat_out_channels[2],
            num_features // 2,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=False,
        )
        self.bn4_2 = nn.BatchNorm2d(num_features // 2, momentum=0.01, affine=True)

        self.daspp_3 = AtrousBlock2d(
            num_features // 2,
            num_features // 4,
            3,
            norm_first=False,
            norm_momentum=0.01,
        )
        self.daspp_6 = AtrousBlock2d(
            num_features // 2 + num_features // 4 + feat_out_channels[2],
            num_features // 4,
            6,
            norm_momentum=0.01,
        )
        self.daspp_12 = AtrousBlock2d(
            num_features + feat_out_channels[2],
            num_features // 4,
            12,
            norm_momentum=0.01,
        )
        self.daspp_18 = AtrousBlock2d(
            num_features + num_features // 4 + feat_out_channels[2],
            num_features // 4,
            18,
            norm_momentum=0.01,
        )
        self.daspp_24 = AtrousBlock2d(
            num_features + num_features // 2 + feat_out_channels[2],
            num_features // 4,
            24,
            norm_momentum=0.01,
        )
        self.daspp_conv = torch.nn.Sequential(
            nn.Conv2d(
                num_features + num_features // 2 + num_features // 4,
                num_features // 4,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=False,
            ),
            nn.ELU(),
        )
        self.reduc8x8 = Reduction1x1(num_features // 4, num_features // 4, self.max_depth)
        self.lpg8x8 = LocalPlanarGuidance(8)

        self.upconv3 = UpConv2d(in_channels=num_features // 4, out_channels=num_features // 4)
        self.bn3 = nn.BatchNorm2d(num_features // 4, momentum=0.01, affine=True)
        self.conv3 = nn.Conv2d(
            num_features // 4 + feat_out_channels[1] + 1,
            num_features // 4,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=False,
        )
        self.reduc4x4 = Reduction1x1(num_features // 4, num_features // 8, self.max_depth)
        self.lpg4x4 = LocalPlanarGuidance(4)

        self.upconv2 = UpConv2d(in_channels=num_features // 4, out_channels=num_features // 8)
        self.bn2 = nn.BatchNorm2d(num_features // 8, momentum=0.01, affine=True)
        self.conv2 = nn.Conv2d(
            num_features // 8 + feat_out_channels[0] + 1,
            num_features // 8,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=False,
        )

        self.reduc2x2 = Reduction1x1(num_features // 8, num_features // 16, self.max_depth)
        self.lpg2x2 = LocalPlanarGuidance(2)

        self.upconv1 = UpConv2d(in_channels=num_features // 8, out_channels=num_features // 16)
        self.reduc1x1 = Reduction1x1(
            num_features // 16,
            num_features // 32,
            self.max_depth,
            is_final=True,
        )
        self.conv1 = nn.Conv2d(num_features // 16 + 4, num_features // 16, 3, 1, 1, bias=False)
        self.output_channels = 1

        if dist_family is not None:
            dist_layer_class = get_dist_conv_layer(dist_family)
            self.depth = dist_layer_class(
                base_layer=nn.Conv2d,
                event_dim=self.output_channels,
                in_channels=num_features // 16,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=False,
            )
        else:
            self.depth = nn.Conv2d(num_features // 16, self.output_channels, 3, 1, 1, bias=False)

    def feat_forward(self, features: list[Tensor]) -> Tensor:
        dense_features = F.relu(features[4])
        upconv5 = self.bn5(self.upconv5(dense_features))  # H/16
        iconv5 = F.elu(self.conv5(torch.cat([upconv5, features[3]], dim=1)))

        upconv4 = self.bn4(self.upconv4(iconv5))  # H/8
        concat4 = torch.cat([upconv4, features[2]], dim=1)
        iconv4 = self.bn4_2(F.elu(self.conv4(concat4)))

        daspp_3 = self.daspp_3(iconv4)
        concat4_2 = torch.cat([concat4, daspp_3], dim=1)
        daspp_6 = self.daspp_6(concat4_2)
        concat4_3 = torch.cat([concat4_2, daspp_6], dim=1)
        daspp_12 = self.daspp_12(concat4_3)
        concat4_4 = torch.cat([concat4_3, daspp_12], dim=1)
        daspp_18 = self.daspp_18(concat4_4)
        daspp_24 = self.daspp_24(torch.cat([concat4_4, daspp_18], dim=1))
        concat4_daspp = torch.cat([iconv4, daspp_3, daspp_6, daspp_12, daspp_18, daspp_24], dim=1)
        daspp_feat = self.daspp_conv(concat4_daspp)

        reduc8x8 = self.reduc8x8(daspp_feat)
        plane_normal_8x8 = reduc8x8[:, :3, :, :]
        plane_normal_8x8 = F.normalize(plane_normal_8x8, p=2, dim=1)
        plane_dist_8x8 = reduc8x8[:, 3, :, :]
        plane_eq_8x8 = torch.cat([plane_normal_8x8, plane_dist_8x8.unsqueeze(1)], 1)
        depth_8x8 = self.lpg8x8(plane_eq_8x8)
        depth_8x8_scaled = depth_8x8.unsqueeze(1) / self.max_depth
        depth_8x8_scaled_ds = F.interpolate(depth_8x8_scaled, scale_factor=0.25, mode="nearest")

        upconv3 = self.bn3(self.upconv3(daspp_feat))  # H/4
        concat3 = torch.cat([upconv3, features[1], depth_8x8_scaled_ds], dim=1)
        iconv3 = F.elu(self.conv3(concat3))

        reduc4x4 = self.reduc4x4(iconv3)
        plane_normal_4x4 = reduc4x4[:, :3, :, :]
        plane_normal_4x4 = F.normalize(plane_normal_4x4, p=2, dim=1)
        plane_dist_4x4 = reduc4x4[:, 3, :, :]
        plane_eq_4x4 = torch.cat([plane_normal_4x4, plane_dist_4x4.unsqueeze(1)], 1)
        depth_4x4 = self.lpg4x4(plane_eq_4x4)
        depth_4x4_scaled = depth_4x4.unsqueeze(1) / self.max_depth
        depth_4x4_scaled_ds = F.interpolate(depth_4x4_scaled, scale_factor=0.5, mode="nearest")

        upconv2 = self.bn2(self.upconv2(iconv3))  # H/2
        iconv2 = F.elu(self.conv2(torch.cat([upconv2, features[0], depth_4x4_scaled_ds], dim=1)))

        reduc2x2 = self.reduc2x2(iconv2)
        plane_normal_2x2 = reduc2x2[:, :3, :, :]
        plane_normal_2x2 = F.normalize(plane_normal_2x2, p=2, dim=1)
        plane_dist_2x2 = reduc2x2[:, 3, :, :]
        plane_eq_2x2 = torch.cat([plane_normal_2x2, plane_dist_2x2.unsqueeze(1)], 1)
        depth_2x2 = self.lpg2x2(plane_eq_2x2)
        depth_2x2_scaled = depth_2x2.unsqueeze(1) / self.max_depth

        upconv1 = self.upconv1(iconv2)
        reduc1x1 = self.reduc1x1(upconv1)
        concat1 = torch.cat(
            [
                upconv1,
                reduc1x1,
                depth_2x2_scaled,
                depth_4x4_scaled,
                depth_8x8_scaled,
            ],
            dim=1,
        )
        return F.elu(self.conv1(concat1))

    def forward(self, features: list[Tensor]) -> Tensor | dict[str, Tensor]:
        """Forward pass.

        Args:
            features (list[Tensor]): list of the features from the backbone.

        Note:
            Depending of the :attr:`dist_family` of the backbone, the output can
            be a dictionnary of distribution parameters or a single tensor.
        """
        # TODO: handle focal
        out = self.depth(self.feat_forward(features))
        if isinstance(out, dict):
            if "loc" not in out or "scale" not in out:
                raise ValueError(
                    "Expected 'loc' and 'scale' in the output dictionary.",
                    "Consider raising an issue on the repository if you need support ",
                    "for distributions that do not use location and scale.",
                )
            out["loc"] = self.max_depth * F.sigmoid(out["loc"])
            out["scale"] = self.max_depth * out["scale"]
        else:
            out = self.max_depth * F.sigmoid(out)
        return out


class _BTS(nn.Module):
    def __init__(
        self,
        backbone_name: Literal[
            "densenet121",
            "densenet161",
            "resnet50",
            "resnet101",
            "resnext50",
            "resnext101",
        ],
        max_depth: float,
        bts_size: int = 512,
        dist_family: str | None = None,
        pretrained_backbone: bool = True,
    ) -> None:
        """BTS model.

        Args:
            backbone_name (str): Name of the encoding backbone.
            max_depth (float): Maximum predicted depth.
            bts_size (int): BTS feature size. Defaults to 512.
            dist_family (str): Distribution family name. Defaults to None.
            estimation. Defaults to nn.Identity.
            pretrained_backbone (bool): Use a pretrained backbone. Defaults to True.

        Reference:
            From Big to Small: Multi-Scale Local Planar Guidance for Monocular Depth Estimation.
            Jin Han Lee, Myung-Kyu Han, Dong Wook Ko, Il Hong Suh. ArXiv.
        """
        super().__init__()
        self.max_depth = max_depth

        self.backbone = BTSBackbone(backbone_name, pretrained_backbone)
        self.decoder = BTSDecoder(max_depth, self.backbone.feat_out_channels, bts_size, dist_family)

    # TODO: Handle focal
    def forward(self, x: Tensor, focal: float | None = None) -> Tensor:
        """Forward pass.

        Args:
            x (Tensor): Input tensor.
            focal (float): Focal length for API consistency.
        """
        return self.decoder(self.backbone(x))


def _bts(
    backbone_name: str,
    max_depth: float,
    bts_size: int = 512,
    dist_family: str | None = None,
    pretrained_backbone: bool = True,
) -> _BTS:
    if backbone_name not in bts_backbones:
        raise ValueError(f"Unsupported backbone. Got {backbone_name}.")
    return _BTS(backbone_name, max_depth, bts_size, dist_family, pretrained_backbone)


def bts_resnet50(
    max_depth: float,
    bts_size: int = 512,
    dist_family: str | None = None,
    pretrained_backbone: bool = True,
) -> _BTS:
    """BTS model with ResNet-50 backbone.

    Args:
        max_depth (float): Maximum predicted depth.
        bts_size (int): BTS feature size. Defaults to 512.
        dist_family (str): Distribution family name. Defaults to None.
        pretrained_backbone (bool): Use a pretrained backbone. Defaults to True.
    """
    return _bts(
        "resnet50",
        max_depth,
        bts_size=bts_size,
        dist_family=dist_family,
        pretrained_backbone=pretrained_backbone,
    )


def bts_resnet101(
    max_depth: float,
    bts_size: int = 512,
    dist_family: str | None = None,
    pretrained_backbone: bool = True,
) -> _BTS:
    """BTS model with ResNet-101 backbone.

    Args:
        max_depth (float): Maximum predicted depth.
        bts_size (int): BTS feature size. Defaults to 512.
        dist_family (str): Distribution family name. Defaults to None.
        pretrained_backbone (bool): Use a pretrained backbone. Defaults to True.
    """
    return _bts(
        "resnet101",
        max_depth,
        bts_size=bts_size,
        dist_family=dist_family,
        pretrained_backbone=pretrained_backbone,
    )



import torch
from einops import repeat
from torch import nn

from torch_uncertainty.layers import BatchConv2d, BatchLinear


class BatchEnsemble(nn.Module):
    def __init__(
        self,
        model: nn.Module,
        num_estimators: int,
        repeat_training_inputs: bool = False,
        convert_layers: bool = False,
    ) -> None:
        """Wrap a BatchEnsemble model to ensure correct batch replication.

        In a BatchEnsemble architecture, each estimator operates on a **sub-batch**
        of the input. This means that the input batch must be **repeated**
        :attr:`num_estimators` times before being processed.

        This wrapper automatically **duplicates the input batch** along the first axis,
        ensuring that each estimator receives the correct data format.

        Args:
            model (nn.Module): The BatchEnsemble model.
            num_estimators (int): Number of ensemble members.
            repeat_training_inputs (optional, bool): Whether to repeat the input batch during training.
                If ``True``, the input batch is repeated during both training and evaluation. If ``False``,
                the input batch is repeated only during evaluation. Default is ``False``.
            convert_layers (optional, bool): Whether to convert the model's layers to BatchEnsemble layers.
                If ``True``, the wrapper will convert all ``nn.Linear`` and ``nn.Conv2d`` layers to their
                BatchEnsemble counterparts. Default is ``False``.

        Raises:
            ValueError: If neither ``BatchLinear`` nor ``BatchConv2d`` layers are found in the model at the
                end of initialization.
            ValueError: If ``num_estimators`` is less than or equal to ``0``.
            ValueError: If ``convert_layers=True`` and neither ``nn.Linear`` nor ``nn.Conv2d`` layers are
                found in the model.

        Warning:
            If ``convert_layers==True``, the wrapper will attempt to convert all ``nn.Linear`` and ``nn.Conv2d``
            layers in the model to their BatchEnsemble counterparts. If the model contains other types of
            layers, the conversion won't happen for these layers. If don't have any ``nn.Linear`` or ``nn.Conv2d``
            layers in the model, the wrapper will raise an error during conversion.

        Warning:
            If ``repeat_training_inputs==True`` and you want to use one of the ``torch_uncertainty.routines``
            for training, be sure to set ``format_batch_fn=RepeatTarget(num_repeats=num_estimators)`` when
            initializing the routine.

        Example:
            >>> model = nn.Sequential(
            ...     nn.Linear(10, 5),
            ...     nn.ReLU(),
            ...     nn.Linear(5, 2)
            ... )
            >>> model = BatchEnsemble(model, num_estimators=4, convert_layers=True)
            >>> model
            BatchEnsemble(
            (model): Sequential(
                (0): BatchLinear(in_features=10, out_features=5, num_estimators=4)
                (1): ReLU()
                (2): BatchLinear(in_features=5, out_features=2, num_estimators=4)
            )
            )
        """
        super().__init__()
        self.model = model
        self.num_estimators = num_estimators
        self.repeat_training_inputs = repeat_training_inputs

        if convert_layers:
            self._convert_layers()

        filtered_modules = [
            module
            for module in self.model.modules()
            if isinstance(module, BatchLinear | BatchConv2d)
        ]
        _batch_ensemble_checks(filtered_modules, num_estimators)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Repeat the input if ``self.training==False`` or ``repeat_training_inputs==True`` and pass it through the model."""
        if not self.training or self.repeat_training_inputs:
            x = repeat(x, "b ... -> (m b) ...", m=self.num_estimators)
        return self.model(x)

    def _convert_layers(self) -> None:
        """Convert the model's layers to BatchEnsemble layers."""
        no_valid_layers = True
        for name, layer in self.model.named_modules():
            if isinstance(layer, nn.Linear):
                setattr(
                    self.model,
                    name,
                    BatchLinear.from_linear(layer, num_estimators=self.num_estimators),
                )
                no_valid_layers = False
            elif isinstance(layer, nn.Conv2d):
                setattr(
                    self.model,
                    name,
                    BatchConv2d.from_conv2d(layer, num_estimators=self.num_estimators),
                )
                no_valid_layers = False
        if no_valid_layers:
            raise ValueError(
                "No valid layers found in the model. "
                "Please use `nn.Linear` or `nn.Conv2d` layers to apply BatchEnsemble."
            )


def _batch_ensemble_checks(filtered_modules, num_estimators):
    """Check if the model contains the required number of dropout modules."""
    if len(filtered_modules) == 0:
        raise ValueError(
            "No BatchEnsemble layers found in the model. "
            "Please use `BatchLinear` or `BatchConv2d` layers in your model "
            "or set `convert_layers=True` when initializing the wrapper."
        )
    if num_estimators <= 0:
        raise ValueError("`num_estimators` must be greater than 0.")


def batch_ensemble(
    model: nn.Module,
    num_estimators: int,
    repeat_training_inputs: bool = False,
    convert_layers: bool = False,
) -> BatchEnsemble:
    """BatchEnsemble wrapper for a model.

    Args:
        model (nn.Module): model to wrap
        num_estimators (int): number of ensemble members
        repeat_training_inputs (bool, optional): whether to repeat the input batch during training.
            If ``True``, the input batch is repeated during both training and evaluation. If ``False``,
            the input batch is repeated only during evaluation. Default is ``False``.
        convert_layers (bool, optional): whether to convert the model's layers to BatchEnsemble layers.
            If ``True``, the wrapper will convert all ``nn.Linear`` and ``nn.Conv2d`` layers to their
            BatchEnsemble counterparts. Default is ``False``.

    Returns:
        BatchEnsemble: BatchEnsemble wrapper for the model
    """
    return BatchEnsemble(
        model=model,
        num_estimators=num_estimators,
        repeat_training_inputs=repeat_training_inputs,
        convert_layers=convert_layers,
    )



from collections.abc import Callable
from functools import partial

import torch
import torch.nn.functional as F
from torch import nn

from torch_uncertainty.layers.bayesian import BayesConv2d, BayesLinear
from torch_uncertainty.layers.mc_batch_norm import MCBatchNorm2d
from torch_uncertainty.layers.packed import PackedConv2d, PackedLinear
from torch_uncertainty.models import StochasticModel
from torch_uncertainty.models.wrappers.batch_ensemble import BatchEnsemble

__all__ = ["bayesian_lenet", "lenet", "packed_lenet"]


class _LeNet(nn.Module):
    def __init__(
        self,
        in_channels: int,
        num_classes: int,
        linear_layer: type[nn.Module],
        conv2d_layer: type[nn.Module],
        layer_args: dict,
        activation: Callable,
        norm: type[nn.Module],
        groups: int,
        dropout_rate: float,
    ) -> None:
        super().__init__()
        self.activation = activation

        batchnorm = False
        if norm == nn.Identity:
            self.norm1 = norm()
            self.norm2 = norm()
        elif norm == nn.BatchNorm2d or (isinstance(norm, partial) and norm.func == MCBatchNorm2d):
            batchnorm = True
        else:
            raise ValueError(f"norm must be nn.Identity or nn.BatchNorm2d. Got {norm}.")

        self.dropout_rate = dropout_rate

        self.conv1 = conv2d_layer(in_channels, 6, (5, 5), groups=groups, **layer_args)
        if batchnorm:
            self.norm1 = norm(6)
        self.conv_dropout = nn.Dropout2d(p=dropout_rate)
        self.conv2 = conv2d_layer(6, 16, (5, 5), groups=groups, **layer_args)
        if batchnorm:
            self.norm2 = norm(16)
        self.pooling = nn.AdaptiveAvgPool2d((4, 4))
        self.fc1 = linear_layer(256, 120, **layer_args)
        self.fc_dropout = nn.Dropout(p=dropout_rate)
        self.fc2 = linear_layer(120, 84, **layer_args)
        self.last_fc_dropout = nn.Dropout(p=dropout_rate)
        self.fc3 = linear_layer(84, num_classes, **layer_args)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = self.conv_dropout(self.activation(self.norm1(self.conv1(x))))
        out = F.max_pool2d(out, 2)
        out = self.conv_dropout(self.activation(self.norm2(self.conv2(out))))
        out = F.max_pool2d(out, 2)
        out = self.pooling(out)
        out = torch.flatten(out, 1)
        out = self.fc_dropout(
            self.activation(self.fc1(out)),
        )
        out = self.last_fc_dropout(self.activation(self.fc2(out)))
        return self.fc3(out)


def _lenet(
    stochastic: bool,
    in_channels: int,
    num_classes: int,
    layer_args: dict,
    num_samples: int = 16,
    linear_layer: type[nn.Module] = nn.Linear,
    conv2d_layer: type[nn.Module] = nn.Conv2d,
    activation: Callable = nn.ReLU,
    norm: type[nn.Module] = nn.Identity,
    groups: int = 1,
    dropout_rate: float = 0.0,
) -> _LeNet | StochasticModel:
    model = _LeNet(
        in_channels=in_channels,
        num_classes=num_classes,
        linear_layer=linear_layer,
        conv2d_layer=conv2d_layer,
        activation=activation,
        norm=norm,
        groups=groups,
        layer_args=layer_args,
        dropout_rate=dropout_rate,
    )
    if stochastic:
        return StochasticModel(model, num_samples)
    return model


def lenet(
    in_channels: int,
    num_classes: int,
    activation: Callable = F.relu,
    norm: type[nn.Module] = nn.Identity,
    groups: int = 1,
    dropout_rate: float = 0.0,
) -> _LeNet:
    return _lenet(
        stochastic=False,
        in_channels=in_channels,
        num_classes=num_classes,
        linear_layer=nn.Linear,
        conv2d_layer=nn.Conv2d,
        layer_args={},
        activation=activation,
        norm=norm,
        groups=groups,
        dropout_rate=dropout_rate,
    )


def batchensemble_lenet(
    in_channels: int,
    num_classes: int,
    num_estimators: int = 4,
    activation: Callable = F.relu,
    norm: type[nn.Module] = nn.BatchNorm2d,
    groups: int = 1,
    dropout_rate: float = 0.0,
    repeat_training_inputs: bool = False,
) -> _LeNet:
    model = lenet(
        in_channels=in_channels,
        num_classes=num_classes,
        activation=activation,
        norm=norm,
        groups=groups,
        dropout_rate=dropout_rate,
    )
    return BatchEnsemble(
        model=model,
        num_estimators=num_estimators,
        repeat_training_inputs=repeat_training_inputs,
        convert_layers=True,
    )


def packed_lenet(
    in_channels: int,
    num_classes: int,
    num_estimators: int = 4,
    alpha: float = 2,
    gamma: float = 1,
    activation: Callable = F.relu,
    norm: type[nn.Module] = nn.Identity,
    groups: int = 1,
    dropout_rate: float = 0.0,
) -> _LeNet:
    return _lenet(
        stochastic=False,
        in_channels=in_channels,
        num_classes=num_classes,
        linear_layer=PackedLinear,
        conv2d_layer=PackedConv2d,
        norm=norm,
        layer_args={
            "num_estimators": num_estimators,
            "alpha": alpha,
            "gamma": gamma,
        },
        activation=activation,
        groups=groups,
        dropout_rate=dropout_rate,
    )


def bayesian_lenet(
    in_channels: int,
    num_classes: int,
    num_samples: int = 16,
    prior_sigma_1: float | None = None,
    prior_sigma_2: float | None = None,
    prior_pi: float | None = None,
    mu_init: float | None = None,
    sigma_init: float | None = None,
    activation: Callable = F.relu,
    norm: type[nn.Module] = nn.Identity,
    groups: int = 1,
    dropout_rate: float = 0.0,
) -> StochasticModel:
    layers_args = {}
    if prior_sigma_1 is not None:
        layers_args["prior_sigma_1"] = prior_sigma_1
    if prior_sigma_2 is not None:
        layers_args["prior_sigma_2"] = prior_sigma_2
    if prior_pi is not None:
        layers_args["prior_pi"] = prior_pi
    if mu_init is not None:
        layers_args["mu_init"] = mu_init
    if sigma_init is not None:
        layers_args["sigma_init"] = sigma_init

    return _lenet(
        stochastic=True,
        num_samples=num_samples,
        in_channels=in_channels,
        num_classes=num_classes,
        linear_layer=BayesLinear,
        conv2d_layer=BayesConv2d,
        norm=norm,
        layer_args=layers_args,
        activation=activation,
        groups=groups,
        dropout_rate=dropout_rate,
    )



from pathlib import Path

from torch.nn.common_types import _size_2_t

from torch_uncertainty.datasets import KITTIDepth

from .base import DepthDataModule


class KITTIDataModule(DepthDataModule):
    def __init__(
        self,
        root: str | Path,
        batch_size: int,
        min_depth: float = 1e-3,
        max_depth: float = 80.0,
        crop_size: _size_2_t = (352, 704),
        eval_size: _size_2_t = (375, 1242),
        val_split: float | None = None,
        num_workers: int = 1,
        pin_memory: bool = True,
        persistent_workers: bool = True,
    ) -> None:
        r"""Depth DataModule for the KITTI-Depth dataset.

        Args:
            root (str or Path): Root directory of the datasets.
            batch_size (int): Number of samples per batch.
            min_depth (float, optional): Minimum depth value for evaluation.
                Defaults to ``1e-3``.
            max_depth (float, optional): Maximum depth value for training and
                evaluation. Defaults to ``80.0``.
            crop_size (sequence or int, optional): Desired input image and
                depth mask sizes during training. If :attr:`crop_size` is an
                int instead of sequence like :math:`(H, W)`, a square crop
                :math:`(\text{size},\text{size})` is made. If provided a sequence
                of length :math:`1`, it will be interpreted as
                :math:`(\text{size[0]},\text{size[1]})`. Defaults to ``(375, 1242)``.
            eval_size (sequence or int, optional): Desired input image and
                depth mask sizes during evaluation. If size is an int,
                smaller edge of the images will be matched to this number, i.e.,
                :math:`\text{height}>\text{width}`, then image will be rescaled to
                :math:`(\text{size}\times\text{height}/\text{width},\text{size})`.
                Defaults to ``(375, 1242)``.
            val_split (float or None, optional): Share of training samples to use
                for validation. Defaults to ``None``.
            num_workers (int, optional): Number of dataloaders to use. Defaults to
                ``1``.
            pin_memory (bool, optional):  Whether to pin memory. Defaults to
                ``True``.
            persistent_workers (bool, optional): Whether to use persistent workers.
                Defaults to ``True``.
        """
        super().__init__(
            dataset=KITTIDepth,
            root=root,
            batch_size=batch_size,
            min_depth=min_depth,
            max_depth=max_depth,
            crop_size=crop_size,
            eval_size=eval_size,
            val_split=val_split,
            num_workers=num_workers,
            pin_memory=pin_memory,
            persistent_workers=persistent_workers,
        )



import pytest

from torch_uncertainty.baselines.classification.deep_ensembles import (
    DeepEnsemblesBaseline,
)


class TestDeepEnsembles:
    """Testing the Deep Ensembles baseline class."""

    def test_failure(self):
        with pytest.raises(ValueError, match="Models must not be an empty list."):
            DeepEnsemblesBaseline(
                log_path=".",
                checkpoint_ids=[],
                backbone="resnet",
                num_classes=10,
            )



# ruff: noqa: F401
from .batch_ensemble import BatchConv2d, BatchLinear
from .bayesian import BayesConv1d, BayesConv2d, BayesConv3d, BayesLinear
from .channel_layer_norm import ChannelLayerNorm
from .masksembles import MaskedConv2d, MaskedLinear
from .modules import Identity
from .packed import (
    PackedConv1d,
    PackedConv2d,
    PackedConv3d,
    PackedLayerNorm,
    PackedLinear,
    PackedMultiheadAttention,
    PackedTransformerDecoderLayer,
    PackedTransformerEncoderLayer,
)



from typing import Literal

from torch import Tensor
from torchmetrics import MeanAbsoluteError, MeanSquaredError


def _unit_to_factor(unit: Literal["mm", "m", "km"]) -> float:
    """Convert a unit to a factor for scaling.

    Args:
        unit: Unit for the computation of the metric. Must be one of 'mm', 'm',
            'km'.
    """
    if unit == "km":
        return 1e-3
    if unit == "m":
        return 1.0
    if unit == "mm":
        return 1e3
    raise ValueError(f"unit must be one of 'mm', 'm', 'km'. Got {unit}.")


class MeanSquaredErrorInverse(MeanSquaredError):
    r"""Mean Squared Error of the inverse predictions (iMSE).

    .. math:: \text{iMSE} = \frac{1}{N}\sum_i^N(\frac{1}{y_i} - \frac{1}{\hat{y_i}})^2

    Where :math:`y` is a tensor of target values, and :math:`\hat{y}` is a
        tensor of predictions.
    Both are scaled by a factor of :attr:`unit_factor` depending on the
        :attr:`unit` given.

    As input to ``forward`` and ``update`` the metric accepts the following
        input:

    - ``preds`` (:class:`~Tensor`): Predictions from model
    - ``target`` (:class:`~Tensor`): Ground truth values

    As output of ``forward`` and ``compute`` the metric returns the following
        output:

    - ``mean_squared_error`` (:class:`~Tensor`): A tensor with the mean
        squared error

    Args:
        squared: If True returns MSE value, if False returns RMSE value.
        num_outputs: Number of outputs in multioutput setting.
        unit: Unit for the computation of the metric. Must be one of 'mm', 'm',
            'km'. Defauts to 'km'.
        kwargs: Additional keyword arguments.
    """

    def __init__(
        self,
        squared: bool = True,
        num_outputs: int = 1,
        unit: str = "km",
        **kwargs,
    ) -> None:
        super().__init__(squared, num_outputs, **kwargs)
        self.unit_factor = _unit_to_factor(unit)

    def update(self, preds: Tensor, target: Tensor) -> None:
        """Update state with predictions and targets."""
        super().update(1 / (preds * self.unit_factor), 1 / (target * self.unit_factor))


class MeanAbsoluteErrorInverse(MeanAbsoluteError):
    r"""Mean Absolute Error of the inverse predictions (iMAE).

    .. math:: \text{iMAE} = \frac{1}{N}\sum_i^N | \frac{1}{y_i} - \frac{1}{\hat{y_i}} |

    Where :math:`y` is a tensor of target values, and :math:`\hat{y}` is a
        tensor of predictions.
    Both are scaled by a factor of :attr:`unit_factor` depending on the
        :attr:`unit` given.

    As input to ``forward`` and ``update`` the metric accepts the following
        input:

    - ``preds`` (:class:`~Tensor`): Predictions from model
    - ``target`` (:class:`~Tensor`): Ground truth values

    As output of ``forward`` and ``compute`` the metric returns the following
        output:

    - ``mean_absolute_inverse_error`` (:class:`~Tensor`): A tensor with the
        mean absolute error over the state

    Args:
        unit: Unit for the computation of the metric. Must be one of 'mm', 'm',
            'km'. Defauts to 'km'.
        kwargs: Additional keyword arguments.
    """

    def __init__(self, unit: str = "km", **kwargs) -> None:
        super().__init__(**kwargs)
        self.unit_factor = _unit_to_factor(unit)

    def update(self, preds: Tensor, target: Tensor) -> None:
        """Update state with predictions and targets."""
        super().update(1 / (preds * self.unit_factor), 1 / (target * self.unit_factor))



from pathlib import Path

from torch_uncertainty.datasets.classification import DOTA2Games

from .uci_classification import UCIClassificationDataModule


class DOTA2GamesDataModule(UCIClassificationDataModule):
    def __init__(
        self,
        root: str | Path,
        batch_size: int,
        val_split: float = 0.0,
        test_split: float = 0.2,
        num_workers: int = 1,
        pin_memory: bool = True,
        persistent_workers: bool = True,
        binary: bool = True,
    ) -> None:
        """The Dota2 Games UCI classification datamodule.

        Args:
            root (string): Root directory of the datasets.
            batch_size (int): The batch size for training and testing.
            val_split (float, optional): Share of validation samples among the
                non-test samples. Defaults to ``0``.
            test_split (float, optional): Share of test samples. Defaults to ``0.2``.
            num_workers (int, optional): How many subprocesses to use for data
                loading. Defaults to ``1``.
            pin_memory (bool, optional): Whether to pin memory in the GPU. Defaults
                to ``True``.
            persistent_workers (bool, optional): Whether to use persistent workers.
                Defaults to ``True``.
            binary (bool, optional): Whether to use binary classification. Defaults
                to ``True``.

        """
        super().__init__(
            root=root,
            dataset=DOTA2Games,
            batch_size=batch_size,
            val_split=val_split,
            test_split=test_split,
            num_workers=num_workers,
            pin_memory=pin_memory,
            persistent_workers=persistent_workers,
            binary=binary,
        )



from importlib import util
from typing import Literal

from torch import Tensor, nn
from torch.utils.data import DataLoader, Dataset

from .abstract import PostProcessing

if util.find_spec("laplace"):
    from laplace import Laplace

    laplace_installed = True
else:  # coverage: ignore
    laplace_installed = False


class LaplaceApprox(PostProcessing):
    def __init__(
        self,
        task: Literal["classification", "regression"],
        model: nn.Module | None = None,
        weight_subset="last_layer",
        hessian_struct="kron",
        pred_type: Literal["glm", "nn"] = "glm",
        link_approx: Literal["mc", "probit", "bridge", "bridge_norm"] = "probit",
        batch_size: int = 256,
        optimize_prior_precision: bool = True,
    ) -> None:
        """Laplace approximation for uncertainty estimation.

        This class is a wrapper of Laplace classes from the laplace-torch library.

        Args:
            task (Literal["classification", "regression"]): task type.
            model (nn.Module): model to be converted.
            weight_subset (str): subset of weights to be considered. Defaults to
                "last_layer".
            hessian_struct (str): structure of the Hessian matrix. Defaults to
                "kron".
            pred_type (Literal["glm", "nn"], optional): type of posterior predictive,
                See the Laplace library for more details. Defaults to "glm".
            link_approx (Literal["mc", "probit", "bridge", "bridge_norm"], optional):
                how to approximate the classification link function for the `'glm'`.
                See the Laplace library for more details. Defaults to "probit".
            batch_size (int, optional): batch size for the Laplace approximation.
                Defaults to 256.
            optimize_prior_precision (bool, optional): whether to optimize the prior
                precision. Defaults to True.

        Reference:
            Daxberger et al. Laplace Redux - Effortless Bayesian Deep Learning. In NeurIPS 2021.
        """
        super().__init__()
        if not laplace_installed:  # coverage: ignore
            raise ImportError(
                "The laplace-torch library is not installed. Please install"
                "torch_uncertainty with the all option:"
                """pip install -U "torch_uncertainty[all]"."""
            )

        self.pred_type = pred_type
        self.link_approx = link_approx
        self.task = task
        self.weight_subset = weight_subset
        self.hessian_struct = hessian_struct
        self.batch_size = batch_size
        self.optimize_prior_precision = optimize_prior_precision

        if model is not None:
            self.set_model(model)

    def set_model(self, model: nn.Module) -> None:
        super().set_model(model)
        self.la = Laplace(
            model=model,
            likelihood=self.task,
            subset_of_weights=self.weight_subset,
            hessian_structure=self.hessian_struct,
        )

    def fit(self, dataset: Dataset) -> None:
        dl = DataLoader(dataset, batch_size=self.batch_size)
        self.la.fit(train_loader=dl)
        if self.optimize_prior_precision:
            self.la.optimize_prior_precision(method="marglik")

    def forward(
        self,
        x: Tensor,
    ) -> Tensor:
        return self.la(x, pred_type=self.pred_type, link_approx=self.link_approx)



import math

import torch
from einops import repeat
from torch import Tensor, nn
from torch.nn.common_types import _size_2_t
from torch.nn.modules.utils import _pair


class BatchLinear(nn.Module):
    __constants__ = ["in_features", "out_features", "num_estimators"]
    in_features: int
    out_features: int
    num_estimators: int
    r_group: Tensor
    s_group: Tensor
    bias: Tensor | None

    def __init__(
        self,
        in_features: int,
        out_features: int,
        num_estimators: int,
        bias: bool = True,
        device=None,
        dtype=None,
    ) -> None:
        r"""BatchEnsemble-style Linear layer.

        Apply a linear transformation using BatchEnsemble method to the incoming
        data.

        .. math::
            y=(x\circ \widehat{r_{group}})W^{T}\circ \widehat{s_{group}} + \widehat{b}

        Args:
            in_features (int): Number of input features..
            out_features (int): Number of output features.
            num_estimators (int): number of estimators in the ensemble, referred as
                :math:`M`.
            bias (bool, optional): if ``True``, adds a learnable bias to the
                output. Defaults to ``True``.
            device (Any, optional): device to use for the parameters and
                buffers of this module. Defaults to ``None``.
            dtype (Any, optional): data type to use for the parameters and
                buffers of this module. Defaults to ``None``.

        Reference:
            Introduced by the paper `BatchEnsemble: An Alternative Approach to
            Efficient Ensemble and Lifelong Learning
            <https://arxiv.org/abs/2002.06715>`_, we present here an implementation
            of a Linear BatchEnsemble layer in `PyTorch <https://pytorch.org>`_
            heavily inspired by its `official implementation
            <https://github.com/google/edward2>`_ in `TensorFlow
            <https://www.tensorflow.org>`_.

        Attributes:
            weight: the learnable weights (:math:`W`) of shape
                :math:`(H_{out}, H_{in})` shared between the estimators. The values
                are initialized from :math:`\mathcal{U}(-\sqrt{k}, \sqrt{k})`,
                where :math:`k = \frac{1}{H_{in}}`.
            r_group: the learnable matrice of shape :math:`(M, H_{in})` where each row
                consist of the vector :math:`r_{i}` corresponding to the
                :math:`i^{th}` ensemble member. The values are initialized from
                :math:`\mathcal{N}(1.0, 0.5)`.
            s_group: the learnable matrice of shape :math:`(M, H_{out})` where each row
                consist of the vector :math:`s_{i}` corresponding to the
                :math:`i^{th}` ensemble member. The values are initialized from
                :math:`\mathcal{N}(1.0, 0.5)`.
            bias: the learnable bias (:math:`b`) of shape :math:`(M, H_{out})`
                where each row corresponds to the bias of the :math:`i^{th}`
                ensemble member. If :attr:`bias` is ``True``, the values are
                initialized from :math:`\mathcal{U}(-\sqrt{k}, \sqrt{k})` where
                :math:`k = \frac{1}{H_{in}}`.

        Shape:
            - Input: :math:`(N, H_{in})` where :math:`N` is the batch size and
              :math:`H_{in} = \text{in_features}`.
            - Output: :math:`(N, H_{out})` where
              :math:`H_{out} = \text{out_features}`.

        Warning:
            It is advised to ensure that `batch_size` is divisible by :attr:`num_estimators` when
            calling :func:`forward()`, so each estimator receives the same number of examples.
            In a BatchEnsemble architecture, the input is typically **repeated** `num_estimators`
            times along the batch dimension. Incorrect batch size may lead to unexpected results.

            To simplify batch handling, wrap your model with `torch_uncertainty.wrappers.BatchEnsemble`,
            which automatically repeats the batch before passing it through the network.


        Examples:
            >>> # With three estimators
            >>> m = BatchLinear(20, 30, 3)
            >>> input = torch.randn(8, 20)
            >>> output = m(input)
            >>> print(output.size())
            torch.Size([8, 30])
        """
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.num_estimators = num_estimators

        self.linear = nn.Linear(
            in_features=in_features,
            out_features=out_features,
            bias=False,
            **factory_kwargs,
        )

        self.r_group = nn.Parameter(torch.empty((num_estimators, in_features), **factory_kwargs))
        self.s_group = nn.Parameter(torch.empty((num_estimators, out_features), **factory_kwargs))
        if bias:
            self.bias = nn.Parameter(torch.empty((num_estimators, out_features), **factory_kwargs))
        else:
            self.register_parameter("bias", None)
        self.reset_parameters()

    @classmethod
    def from_linear(cls, linear: nn.Linear, num_estimators: int) -> "BatchLinear":
        r"""Create a BatchEnsemble-style Linear layer from an existing Linear layer.

        Args:
            linear (nn.Linear): The Linear layer to convert.
            num_estimators (int): Number of ensemble members.

        Returns:
            BatchLinear: The converted BatchEnsemble-style Linear layer.

        Example:
            >>> linear = nn.Linear(20, 30)
            >>> be_linear = BatchLinear.from_linear(linear, num_estimators=3)
        """
        return cls(
            in_features=linear.in_features,
            out_features=linear.out_features,
            num_estimators=num_estimators,
            bias=linear.bias is not None,
            device=linear.weight.device,
            dtype=linear.weight.dtype,
        )

    def reset_parameters(self) -> None:
        nn.init.normal_(self.r_group, mean=1.0, std=0.5)
        nn.init.normal_(self.s_group, mean=1.0, std=0.5)
        if self.bias is not None:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.linear.weight)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            nn.init.uniform_(self.bias, -bound, bound)

    def forward(self, inputs: Tensor) -> Tensor:
        batch_size = inputs.size(0)
        examples_per_estimator = torch.tensor(
            batch_size // self.num_estimators, device=inputs.device
        )
        extra = batch_size % self.num_estimators

        r_group = repeat(self.r_group, "m h -> (m b) h", b=examples_per_estimator)
        r_group = torch.cat([r_group, r_group[:extra]], dim=0)
        s_group = repeat(self.s_group, "m h -> (m b) h", b=examples_per_estimator)
        s_group = torch.cat([s_group, s_group[:extra]], dim=0)
        if self.bias is not None:
            bias = repeat(self.bias, "m h -> (m b) h", b=examples_per_estimator)
            bias = torch.cat([bias, bias[:extra]], dim=0)
        else:
            bias = None

        return self.linear(inputs * r_group) * s_group + (bias if bias is not None else 0)

    def extra_repr(self) -> str:
        return (
            f"in_features={self.in_features},"
            f" out_features={self.out_features},"
            f" num_estimators={self.num_estimators},"
            f" bias={self.bias is not None}"
        )


class BatchConv2d(nn.Module):
    __constants__ = [
        "stride",
        "padding",
        "dilation",
        "groups",
        "in_channels",
        "out_channels",
        "kernel_size",
        "num_estimators",
    ]
    in_channels: int
    out_channels: int
    kernel_size: tuple[int, ...]
    num_estimators: int
    stride: tuple[int, ...]
    padding: str | tuple[int, ...]
    dilation: tuple[int, ...]
    groups: int
    weight: Tensor
    r_group: Tensor
    s_group: Tensor
    bias: Tensor | None

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_2_t,
        num_estimators: int,
        stride: _size_2_t = 1,
        padding: str | _size_2_t = 0,
        dilation: _size_2_t = 1,
        groups: int = 1,
        bias: bool = True,
        device=None,
        dtype=None,
    ) -> None:
        r"""BatchEnsemble-style Conv2d layer.
        
        Applies a 2d convolution over an input signal composed of several input
        planes using BatchEnsemble method to the incoming data.

        In the simplest case, the output value of the layer with input size
        :math:`(N, C_{in}, H_{in}, W_{in})` and output
        :math:`(N, C_{out}, H_{out}, W_{out})` can be precisely described as:

        .. math::
            \text{out}(N_i, C_{\text{out}_j})=\
            &\widehat{b}(N_i,C_{\text{out}_j})
            +\widehat{s_{group}}(N_{i},C_{\text{out}_j}) \\
            &\times \sum_{k = 0}^{C_{\text{in}} - 1}
            \text{weight}(C_{\text{out}_j}, k)\star (\text{input}(N_i, k)
            \times \widehat{r_{group}}(N_i, k))

        Reference:
            Introduced by the paper `BatchEnsemble: An Alternative Approach to
            Efficient Ensemble and Lifelong Learning
            <https://arxiv.org/abs/2002.06715>`_, we present here an implementation
            of a Conv2d BatchEnsemble layer in `PyTorch <https://pytorch.org>`_
            heavily inspired by its `official implementation
            <https://github.com/google/edward2>`_ in `TensorFlow
            <https://www.tensorflow.org>`_.

        Args:
            in_channels (int): number of channels in the input images.
            out_channels (int): number of channels produced by the convolution.
            kernel_size (int or tuple): size of the convolving kernel.
            num_estimators (int): number of estimators in the ensemble referred as
                :math:`M` here.
            stride (int or tuple, optional): stride of the convolution. Defaults to
                ``1``.
            padding (int, tuple or str, optional): padding added to all four sides
                of the input. Defaults to ``0``.
            dilation (int or tuple, optional): spacing between kernel elements.
                Defaults to ``1``.
            groups (int, optional): number of blocked connections from input
                channels to output channels. Defaults to ``1``.
            bias (bool, optional): if ``True``, adds a learnable bias to the
                output. Defaults to ``True``.
            device (Any, optional): device to use for the parameters and
                buffers of this module. Defaults to ``None``.
            dtype (Any, optional): data type to use for the parameters and
                buffers of this module. Defaults to ``None``.

        Attributes:
            weight: the learnable weights of the module of shape
                :math:`(\text{out_channels}, \frac{\text{in_channels}}
                {\text{groups}},`:math:`\text{kernel_size[0]},
                \text{kernel_size[1]})` shared between the estimators. The values
                of these weights are sampled from :math:`\mathcal{U}(-\sqrt{k},
                \sqrt{k})` where :math:`k = \frac{\text{groups}}{C_\text{in} *
                \prod_{i=0}^{1}\text{kernel_size}[i]}`.
            r_group: the learnable matrice of shape :math:`(M, C_{in})` where each row
                consist of the vector :math:`r_{i}` corresponding to the
                :math:`i^{th}` ensemble member. The values are initialized from
                :math:`\mathcal{N}(1.0, 0.5)`.
            s_group: the learnable matrice of shape :math:`(M, C_{out})` where each row
                consist of the vector :math:`s_{i}` corresponding to the
                :math:`i^{th}` ensemble member. The values are initialized from
                :math:`\mathcal{N}(1.0, 0.5)`.
            bias: the learnable bias (:math:`b`) of shape :math:`(M, C_{out})`
                where each row corresponds to the bias of the :math:`i^{th}`
                ensemble member. If :attr:`bias` is ``True``, the values are
                initialized from :math:`\mathcal{U}(-\sqrt{k}, \sqrt{k})` where
                :math:`k=\frac{\text{groups}}{C_\text{in}*\prod_{i=0}^{1}
                \text{kernel_size}[i]}`.

        Shape:
            - Input: :math:`(N, C_{in}, H_{in}, W_{in})`.
            - Output: :math:`(N, C_{out}, H_{out}, W_{out})`.

            .. math::
                H_{out} = \left\lfloor\frac{H_{in} + 2 \times \text{padding}[0] -
                \text{dilation}[0] \times (\text{kernel_size}[0] - 1) - 1}
                {\text{stride}[0]} + 1\right\rfloor

            .. math::
                W_{out} = \left\lfloor\frac{W_{in}  + 2 \times \text{padding}[1] -
                \text{dilation}[1] \times (\text{kernel_size}[1] - 1) - 1}
                {\text{stride}[1]} + 1\right\rfloor

        Warning:
            Ensure that `batch_size` is divisible by :attr:`num_estimators` when calling :func:`forward()`.
            In a BatchEnsemble architecture, the input batch is typically **repeated** `num_estimators`
            times along the first axis. Incorrect batch size may lead to unexpected results.

            To simplify batch handling, wrap your model with `BatchEnsembleWrapper`, which automatically
            repeats the batch before passing it through the network. See `BatchEnsembleWrapper` for details.

        Examples:
            >>> # With square kernels, four estimators and equal stride
            >>> m = BatchConv2d(3, 32, 3, 4, stride=1)
            >>> input = torch.randn(8, 3, 16, 16)
            >>> output = m(input)
            >>> print(output.size())
            torch.Size([8, 32, 14, 14])
        """
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = _pair(kernel_size)
        self.num_estimators = num_estimators
        self.stride = _pair(stride)
        self.padding = padding if isinstance(padding, str) else _pair(padding)
        self.dilation = _pair(dilation)

        self.conv = nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=False,
            **factory_kwargs,
        )
        self.r_group = nn.Parameter(torch.empty((num_estimators, in_channels), **factory_kwargs))
        self.s_group = nn.Parameter(torch.empty((num_estimators, out_channels), **factory_kwargs))
        if bias:
            self.bias = nn.Parameter(torch.empty((num_estimators, out_channels), **factory_kwargs))
        else:
            self.register_parameter("bias", None)

        self.reset_parameters()

    @classmethod
    def from_conv2d(cls, conv2d: nn.Conv2d, num_estimators: int) -> "BatchConv2d":
        r"""Create a BatchEnsemble-style Conv2d layer from an existing Conv2d layer.

        Args:
            conv2d (nn.Conv2d): The Conv2d layer to convert.
            num_estimators (int): Number of ensemble members.

        Returns:
            BatchConv2d: The converted BatchEnsemble-style Conv2d layer.

        Warning:
            All parameters of the original Conv2d layer will be discarded.

        Example:
            >>> conv2d = nn.Conv2d(3, 32, kernel_size=3)
            >>> be_conv2d = BatchConv2d.from_conv2d(conv2d, num_estimators=3)
        """
        return cls(
            in_channels=conv2d.in_channels,
            out_channels=conv2d.out_channels,
            kernel_size=conv2d.kernel_size,
            stride=conv2d.stride,
            padding=conv2d.padding,
            dilation=conv2d.dilation,
            groups=conv2d.groups,
            bias=conv2d.bias is not None,
            num_estimators=num_estimators,
            device=conv2d.weight.device,
            dtype=conv2d.weight.dtype,
        )

    def reset_parameters(self) -> None:
        nn.init.normal_(self.r_group, mean=1.0, std=0.5)
        nn.init.normal_(self.s_group, mean=1.0, std=0.5)
        if self.bias is not None:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.conv.weight)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            nn.init.uniform_(self.bias, -bound, bound)

    def forward(self, inputs: Tensor) -> Tensor:
        batch_size = inputs.size(0)
        examples_per_estimator = batch_size // self.num_estimators
        extra = batch_size % self.num_estimators

        r_group = repeat(self.r_group, "m h -> (m b) h 1 1", b=examples_per_estimator)
        r_group = torch.cat([r_group, r_group[:extra]], dim=0)
        s_group = repeat(self.s_group, "m h -> (m b) h 1 1", b=examples_per_estimator)
        s_group = torch.cat([s_group, s_group[:extra]], dim=0)

        if self.bias is not None:
            bias = repeat(self.bias, "m h -> (m b) h 1 1", b=examples_per_estimator)
            bias = torch.cat([bias, bias[:extra]], dim=0)
        else:
            bias = None

        return self.conv(inputs * r_group) * s_group + (bias if bias is not None else 0)

    def extra_repr(self) -> str:
        return (
            f"in_channels={self.in_channels},"
            f" out_channels={self.out_channels},"
            f" kernel_size={self.kernel_size},"
            f" num_estimators={self.num_estimators},"
            f" stride={self.stride}"
        )



import logging
from abc import ABC, abstractmethod
from collections.abc import Callable
from pathlib import Path

import torch
from torch import Generator
from torch.utils.data import Dataset
from torchvision.datasets.utils import (
    check_integrity,
    download_and_extract_archive,
)


class UCIClassificationDataset(ABC, Dataset):
    """The UCI classification dataset base class.

    Args:
        root (str): Root directory of the datasets.
        train (bool, optional): If True, creates dataset from training set,
            otherwise creates from test set.
        transform (callable, optional): A function/transform that takes in a
            numpy array and returns a transformed version.
        target_transform (callable, optional): A function/transform that takes
            in the target and transforms it.
        download (bool, optional): If true, downloads the dataset from the
            internet and puts it in root directory. If dataset is already
            downloaded, it is not downloaded again.
        binary (bool, optional): Whether to use binary classification. Defaults
            to ``True``.

    Note - License:
        The licenses of the datasets may differ from TorchUncertainty's
        license. Check before use.

    """

    md5_zip: str = ""
    url: str = ""
    filename: str = ""
    dataset_name: str = ""
    need_split = True
    apply_standardization = True

    def __init__(
        self,
        root: Path | str,
        transform: Callable | None = None,
        target_transform: Callable | None = None,
        binary: bool = True,
        download: bool = False,
        train: bool = True,
        test_split: float = 0.2,
        split_seed: int = 21893027,
    ) -> None:
        super().__init__()
        self.root = Path(root)
        self.train = train
        self.transform = transform
        self.target_transform = target_transform

        if download:
            self.download()

        self._make_dataset()
        if self.apply_standardization:
            self._compute_statistics()
            self._standardize()

        if self.need_split:
            gen = Generator().manual_seed(split_seed)

            self.split_idx = torch.ones(len(self)).multinomial(
                num_samples=int((1 - test_split) * len(self)),
                replacement=False,
                generator=gen,
            )
            if not self.train:
                self.split_idx = torch.tensor(
                    [i for i in range(len(self)) if i not in self.split_idx]
                )
            self.data = self.data[self.split_idx]
            self.targets = self.targets[self.split_idx]
        if not binary:
            self.targets = torch.nn.functional.one_hot(self.targets, num_classes=2)

    def __len__(self) -> int:
        """Get the length of the dataset."""
        return self.data.shape[0]

    def _check_integrity(self) -> bool:
        """Check the integrity of the dataset(s)."""
        return check_integrity(
            self.root / Path(self.dataset_name + ".zip"),
            self.md5_zip,
        )

    def _standardize(self) -> None:
        self.data = (self.data - self.data_mean) / self.data_std

    def _compute_statistics(self) -> None:
        self.data_mean = self.data.mean(dim=0)
        self.data_std = self.data.std(dim=0)
        self.data_std[self.data_std == 0] = 1

    def download(self) -> None:
        """Download and extract dataset."""
        if self._check_integrity():
            logging.info("Files already downloaded and verified")
            return
        download_and_extract_archive(
            self.url,
            download_root=self.root / self.dataset_name,
            filename=self.filename + ".zip",
            md5=self.md5_zip,
        )

    @abstractmethod
    def _make_dataset(self) -> None:
        """Create dataset from extracted files."""

    def __getitem__(self, index: int) -> tuple[torch.Tensor, torch.Tensor]:
        """Get sample and target for a given index."""
        data = self.data[index, :]
        if self.transform is not None:
            data = self.transform(data)
        target = self.targets[index]
        if self.target_transform is not None:
            target = self.target_transform(target)
        return data, target



import torch
from torch import Tensor
from torch.nn import LayerNorm

from .utils import ChannelBack, ChannelFront


class ChannelLayerNorm(LayerNorm):
    def __init__(
        self,
        normalized_shape: int | list[int],
        eps: float = 0.00001,
        elementwise_affine: bool = True,
        bias: bool = True,
        device: torch.device | str | None = None,
        dtype: torch.dtype | str | None = None,
    ) -> None:
        r"""Layer normalization over the channel dimension.

        Args:
        normalized_shape (int or list or torch.Size): input shape from an expected input
            of size

            .. math::
                [* \times \text{normalized\_shape}[0] \times \text{normalized\_shape}[1]
                    \times \ldots \times \text{normalized\_shape}[-1]]

            If a single integer is used, it is treated as a singleton list, and this module will
            normalize over the channel dimension which is expected to be of that specific size.
        eps (float): a value added to the denominator for numerical stability. Default: 1e-5
        elementwise_affine (bool): a boolean value that when set to ``True``, this module
            has learnable per-element affine parameters initialized to ones (for weights)
            and zeros (for biases). Default: ``True``.
        bias (bool): If set to ``False``, the layer will not learn an additive bias (only relevant if
            :attr:`elementwise_affine` is ``True``). Default: ``True``.
        device (torch.device or str or None): the desired device of the module.
        dtype (torch.dtype or str or None): the desired floating point type of the module.

        Attributes:
            weight: the learnable weights of the module of shape
                :math:`\text{normalized\_shape}` when :attr:`elementwise_affine` is set to ``True``.
                The values are initialized to 1.
            bias:   the learnable bias of the module of shape
                    :math:`\text{normalized\_shape}` when :attr:`elementwise_affine` is set to ``True``.
                    The values are initialized to 0.

        Shape:
            - Input: :math:`(N, *)`
            - Output: :math:`(N, *)` (same shape as input)

        """
        super().__init__(normalized_shape, eps, elementwise_affine, bias, device, dtype)
        self.cback = ChannelBack()
        self.cfront = ChannelFront()

    def forward(self, inputs: Tensor) -> Tensor:
        return self.cfront(super().forward(self.cback(inputs)))



import logging
from collections.abc import Callable
from pathlib import Path
from typing import Any, Literal

from torchvision.datasets import ImageFolder
from torchvision.datasets.utils import (
    check_integrity,
    download_and_extract_archive,
)


class NotMNIST(ImageFolder):
    """The notMNIST dataset.

    Args:
        root (str): Root directory of the datasets.
        subset (str): The subset to use, one of ``small`` or ``large``.
        transform (callable, optional): A function/transform that takes in
            a PIL image and returns a transformed version. E.g,
            ``transforms.RandomCrop``. Defaults to None.
        target_transform (callable, optional): A function/transform that
            takes in the target and transforms it. Defaults to None.
        download (bool, optional): If True, downloads the dataset from the
            internet and puts it in root directory. If dataset is already
            downloaded, it is not downloaded again. Defaults to False.

    Note:
        There is no information on the license of the dataset. It may not
        be suitable for commercial use.
    """

    url_base = "https://zenodo.org/record/8274268/files/"
    filenames = ["notMNIST_small.zip", "notMNIST_large.zip"]
    tgz_md5s = [
        "3de91fb69221d9c2d5c57387101ebc6c",
        "c3f9e0862df000a897766593044e366a",
    ]
    subsets = ["small", "large"]

    def __init__(
        self,
        root: str | Path,
        subset: Literal["small", "large"] = "small",
        transform: Callable[..., Any] | None = None,
        target_transform: Callable[..., Any] | None = None,
        download: bool = False,
    ) -> None:
        self.root = Path(root)

        if subset not in self.subsets:
            raise ValueError(f"The subset '{subset}' does not exist for notMNIST.")
        ind = self.subsets.index(subset)
        self.url = self.url_base + "/" + self.filenames[ind]
        self.filename = self.filenames[ind]
        self.tgz_md5 = self.tgz_md5s[ind]

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError(
                "Dataset not found or corrupted. You can use download=True to download it."
            )

        super().__init__(
            self.root / f"notMNIST_{subset}",
            transform=transform,
            target_transform=target_transform,
        )

    def _check_integrity(self) -> bool:
        fpath = self.root / self.filename
        return check_integrity(
            fpath,
            self.tgz_md5,
        )

    def download(self) -> None:
        if self._check_integrity():
            logging.info("Files already downloaded and verified")
            return

        download_and_extract_archive(
            self.url,
            download_root=self.root,
            filename=self.filename,
            md5=self.tgz_md5,
        )
        logging.info("Downloaded %s to %s.", self.filename, self.root)



import os
from collections import defaultdict
from collections.abc import Callable
from pathlib import Path
from typing import Literal

import numpy as np
import torch
from PIL import Image
from torch.utils.data import Dataset


class TinyImageNet(Dataset):
    """Inspired by
    https://gist.github.com/z-a-f/b862013c0dc2b540cf96a123a6766e54.
    """

    def __init__(
        self,
        root: str | Path,
        split: Literal["train", "val", "test"] = "train",
        transform: Callable | None = None,
        target_transform: Callable | None = None,
    ) -> None:
        self.root = Path(root) / "tiny-imagenet-200"

        if split not in ["train", "val", "test"]:
            raise ValueError(f"Split {split} is not supported.")

        self.split = split
        self.label_idx = 1
        self.transform = transform
        self.target_transform = target_transform

        self.wnids_path = self.root / "wnids.txt"
        self.words_path = self.root / "words.txt"

        self.make_dataset()

    def make_dataset(self) -> None:
        self.samples_paths = self._make_paths()
        self.samples_num = len(self.samples_paths)

        labels = []
        samples = []
        for idx in range(self.samples_num):
            s = self.samples_paths[idx]
            img = Image.open(s[0])
            img = self._add_channels(np.uint8(img))
            img = Image.fromarray(img)
            samples.append(img)
            labels.append(s[self.label_idx])

        self.samples = samples
        self.label_data = torch.as_tensor(labels).long()

    def _add_channels(self, img: np.ndarray) -> np.ndarray:
        while len(img.shape) < 3:  # third axis is the channels
            img = np.expand_dims(img, axis=-1)
        while (img.shape[-1]) < 3:
            img = np.concatenate([img, img[:, :, -1:]], axis=-1)
        return img

    def __len__(self) -> int:
        """The number of samples in the dataset."""
        return self.samples_num

    def __getitem__(self, index: int) -> tuple[torch.Tensor, torch.Tensor]:
        """Get the samples and targets of the dataset.

        Args:
            index (int): The index of the sample to get.
        """
        sample = self.samples[index]
        target = self.label_data[index]

        if self.transform is not None:
            sample = self.transform(sample)
        if self.target_transform is not None:
            target = self.target_transform(target)

        return sample, target

    def _make_paths(self) -> list[tuple[Path, int]]:
        self.ids = []
        with self.wnids_path.open() as idf:
            for nid in idf:
                snid = nid.strip()
                self.ids.append(snid)
        self.nid_to_words = defaultdict(list)

        with self.words_path.open() as wf:
            for line in wf:
                nid, labels = line.split("\t")
                labels = [x.strip() for x in labels.split(",")]
                self.nid_to_words[nid].extend(labels)

        paths = []

        if self.split == "train":
            train_path = self.root / "train"
            train_nids = os.listdir(train_path)
            for nid in train_nids:
                anno_path = train_path / nid / (nid + "_boxes.txt")
                imgs_path = train_path / nid / "images"
                label_id = self.ids.index(nid)
                with anno_path.open() as annof:
                    for line in annof:
                        fname, _, _, _, _ = line.split()
                        fname = imgs_path / fname
                        paths.append((fname, label_id))

        elif self.split == "val":
            val_path = self.root / "val"
            with (val_path / "val_annotations.txt").open() as valf:
                for line in valf:
                    fname, nid, _, _, _, _ = line.split()
                    fname = val_path / "images" / fname
                    label_id = self.ids.index(nid)
                    paths.append((fname, label_id))

        else:  # self.split == "test":
            test_path = self.root / "test"
            paths = [test_path / x for x in os.listdir(test_path)]
        return paths
